"""zyx.targets"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Generic,
    Type,
    TypeAlias,
    TypeVar,
    TYPE_CHECKING,
)

if TYPE_CHECKING:
    from .operations.validate import ValidationResult
    from ._aliases import (
        PydanticAIInstructions,
        PydanticAIModelSettings,
        PydanticAIUsageLimits,
    )
    from ._types import SourceParam, ModelParam, ContextType, ToolType

__all__ = (
    "Target",
    "TargetHook",
    "target",
)


Deps = TypeVar("Deps")
Output = TypeVar("Output")


TargetHook: TypeAlias = Literal["complete", "error"]


@dataclass
class Target(Generic[Output]):
    """
    A structured way to define the `target` or output to generate
    through a semantic operation.
    """

    target: Output | Type[Output]
    """The target type or value that will be generated/mutated by
    semantic operations."""

    name: str | None = field(default=None)
    """An optional human readable name for this target type, this is useful
    for targets of primitive types as well as for better model understanding
    of a target."""

    description: str | None = field(default=None)
    """An optional human readable description of this target type."""

    instructions: str | None = field(default=None)
    """Specific instructions for the model to follow when generating this target."""

    constraints: List[str] | None = field(default=None)
    """Optional list of constraints the parsed value must satisfy (used by parse/aparse)."""

    model: ModelParam | None = field(default=None)
    """Default model/pydantic ai agent to use for operations on this target (e.g. ``validate``)."""

    constraints: str | List[str] | None = field(default=None)
    """An optional list of constraints that this target type must adhere to,
    these can be explicitly validated by a model if needed."""

    _field_hooks: Dict[str, List[tuple[Callable[[Any], Any], bool, bool]]] = (
        field(init=False, default_factory=dict)
    )
    """A dictionary of field hooks that can be used to validate/mutate this target
    as it is being generated/mutated.

    You can set a field hook on a target by using the `@on_field` decorator."""

    _prebuilt_hooks: Dict[
        TargetHook, List[tuple[Callable[[Any], Any], bool, bool]]
    ] = field(init=False, default_factory=dict)
    """A list of prebuilt hooks that can be used to validate/mutate this target
    based on a certain event or condition.

    You can apply a prebuilt hook on a target by using the `@on` decorator."""

    def on_field(
        self,
        field: str | None = None,
        retry: bool = True,
        update: bool = False,
    ):
        """
        Decorator to register a 'field hook' on this target. A field hook
        is a function that can be used to validate/mutate a specific field
        on the event it is generated or mutated by a model.

        For targets with types such as `str`, `int`, with no field names, you
        can leave the `field` argument as None.
        """

        def decorator(fn: Callable[[Any], Any]) -> Callable[[Any], Any]:
            key = field or "__self__"
            self._field_hooks.setdefault(key, []).append((fn, retry, update))
            return fn

        return decorator

    def on(
        self,
        event: TargetHook,
        retry: bool = True,
        update: bool = False,
    ):
        """
        Decorator to register a 'prebuilt hook' on this target.

        Hooks include:
        - `complete`: Called when the target is generated/mutated successfully.
        - `error`: Called when the target is generated/mutated with an error.
        """

        def decorator(fn: Callable[[Any], Any]) -> Callable[[Any], Any]:
            self._prebuilt_hooks.setdefault(event, []).append(
                (fn, retry, update)
            )
            return fn

        return decorator

    def validate(
        self,
        source: SourceParam,
        *,
        context: ContextType | List[ContextType] | None = None,
        constraints: List[str] | None = None,
        raise_on_error: bool = True,
        model: ModelParam | None = None,
        model_settings: PydanticAIModelSettings | None = None,
        instructions: PydanticAIInstructions | None = None,
        tools: ToolType | List[ToolType] | None = None,
        deps: Deps | None = None,
        usage_limits: PydanticAIUsageLimits | None = None,
    ) -> Output | ValidationResult[Output]:
        """
        Validate a given input `source` against the constraints (or any additional
        constraints) defined on this target, using a `pydantic_ai` model or agent.

        Args:
            source : SourceParam
                The source value to validate against the constraints.
            context : ContextType | List[ContextType] | None = None
                Additional context or conversation history to use for the validation.
            constraints : List[str] | None = None
                Optional list of constraints the parsed value must satisfy (used by parse/aparse).
            raise_on_error : bool = True
                If `True`, raise an error if the constraints are not satisfied.
            model : ModelParam | None = None
                The model or agent to use for the validation.
            model_settings : PydanticAIModelSettings | None = None
                Model settings to use for the validation.
            instructions : PydanticAIInstructions | None = None
                Specific instructions for the model to follow when validating this target.
            tools : ToolType | List[ToolType] | None = None
                Tools to use for the validation.
            deps : Deps | None = None

        Returns:
            Output | ValidationResult[Output]
                The validated output or a `ValidationResult` object if `raise_on_error` is `False`.
        """
        from .operations.validate import validate, ValidationResult

        if not model and not self.model:
            raise ValueError(
                "No model/default model provided for validation. Please either set the `model` attribute on"
                "this target or provide a model to the `validate` method."
            )
        _model = model if model is not None else self.model

        output = validate(
            source=source,
            target=self,
            context=context,
            constraints=constraints,
            raise_on_error=raise_on_error,
            model=_model,  # type: ignore[arg-type]
            model_settings=model_settings,
            instructions=instructions,
            tools=tools,
            deps=deps,
            usage_limits=usage_limits,
        )
        if isinstance(output, ValidationResult):
            return output  # type: ignore[return-value]
        return output.output  # type: ignore[return-value]

    async def avalidate(
        self,
        source: SourceParam,
        *,
        context: ContextType | List[ContextType] | None = None,
        constraints: List[str] | None = None,
        raise_on_error: bool = True,
        model: ModelParam | None = None,
        model_settings: PydanticAIModelSettings | None = None,
        instructions: PydanticAIInstructions | None = None,
        tools: ToolType | List[ToolType] | None = None,
        deps: Deps | None = None,
        usage_limits: PydanticAIUsageLimits | None = None,
    ) -> Output | ValidationResult[Output]:
        """
        Validate a given input `source` against the constraints (or any additional
        constraints) defined on this target, using a `pydantic_ai` model or agent.

        Args:
            source : SourceParam
                The source value to validate against the constraints.
            context : ContextType | List[ContextType] | None = None
                Additional context or conversation history to use for the validation.
            constraints : List[str] | None = None
                Optional list of constraints the parsed value must satisfy (used by parse/aparse).
            raise_on_error : bool = True
                If `True`, raise an error if the constraints are not satisfied.
            model : ModelParam | None = None
                The model or agent to use for the validation.
            model_settings : PydanticAIModelSettings | None = None
                Model settings to use for the validation.
            instructions : PydanticAIInstructions | None = None
                Specific instructions for the model to follow when validating this target.
            tools : ToolType | List[ToolType] | None = None
                Tools to use for the validation.
            deps : Deps | None = None

        Returns:
            Output | ValidationResult[Output]
                The validated output or a `ValidationResult` object if `raise_on_error` is `False`.
        """
        from .operations.validate import avalidate, ValidationResult

        if not model and not self.model:
            raise ValueError(
                "No model/default model provided for validation. Please either set the `model` attribute on"
                "this target or provide a model to the `validate` method."
            )
        _model = model if model is not None else self.model

        output = await avalidate(
            source=source,
            target=self,
            context=context,
            constraints=constraints,
            raise_on_error=raise_on_error,
            model=_model,  # type: ignore[arg-type]
            model_settings=model_settings,
            instructions=instructions,
            tools=tools,
            deps=deps,
            usage_limits=usage_limits,
        )
        if isinstance(output, ValidationResult):
            return output  # type: ignore[return-value]
        return output.output  # type: ignore[return-value]

    def __call__(
        self,
        source: SourceParam,
        *,
        context: ContextType | List[ContextType] | None = None,
        constraints: List[str] | None = None,
        raise_on_error: bool = True,
        model: ModelParam | None = None,
        model_settings: PydanticAIModelSettings | None = None,
        instructions: PydanticAIInstructions | None = None,
        tools: ToolType | List[ToolType] | None = None,
        deps: Deps | None = None,
        usage_limits: PydanticAIUsageLimits | None = None,
    ) -> Output | ValidationResult[Output]:
        """
        Validate a given input `source` against the constraints (or any additional
        constraints) defined on this target, using a `pydantic_ai` model or agent.

        Args:
            source : SourceParam
                The source value to validate against the constraints.
            context : ContextType | List[ContextType] | None = None
                Additional context or conversation history to use for the validation.
            constraints : List[str] | None = None
                Optional list of constraints the parsed value must satisfy (used by parse/aparse).
            raise_on_error : bool = True
                If `True`, raise an error if the constraints are not satisfied.
            model : ModelParam | None = None
                The model or agent to use for the validation.
            model_settings : PydanticAIModelSettings | None = None
                Model settings to use for the validation.
            instructions : PydanticAIInstructions | None = None
                Specific instructions for the model to follow when validating this target.
            tools : ToolType | List[ToolType] | None = None
                Tools to use for the validation.
            deps : Deps | None = None
                Additional dependencies to use for the validation.

        Returns:
            Output | ValidationResult[Output]
                The validated output or a `ValidationResult` object if `raise_on_error` is `False`.
        """
        return self.validate(
            source=source,
            context=context,
            constraints=constraints,
            raise_on_error=raise_on_error,
            model=model,
            model_settings=model_settings,
            instructions=instructions,
            tools=tools,
            deps=deps,
            usage_limits=usage_limits,
        )


def target(
    target: Output | Type[Output],
    name: str | None = None,
    description: str | None = None,
    instructions: str | None = None,
    constraints: List[str] | None = None,
    model: ModelParam | None = None,
) -> Target[Output]:
    """
    A decorator/function to create a new `Target` object from a
    given type or value.

    Args:
        target: Output | Type[Output]
            The target type or value to create a `Target` object from.
        name: str | None = None
            An optional human readable name for this target type, this is useful
            for targets of primitive types as well as for better model understanding
            of a target.
        description: str | None = None
            An optional human readable description of this target type.
        instructions: str | None = None
            Specific instructions for the model to follow when generating this target.
        constraints: List[str] | None = None
            Optional list of constraints the parsed value must satisfy (used by parse/aparse).
        model: ModelParam | None = None
            Default model/pydantic ai agent to use for operations on this target (e.g. ``validate``).

    Returns:
        Target[Output]
            A new `Target` object from the given type or value.
    """
    return Target(
        target=target,
        name=name,
        description=description,
        instructions=instructions,
        constraints=constraints,
        model=model,
    )



"""zyx._utils._confidence

NOTE: This is a re-implementation of the ideas provided by Ruthvik Bandari's
pull request implementing confidence scoring in `instructor`. You can view
the original pull request here:

[PR](https://github.com/567-labs/instructor/pull/1968)
"""

from __future__ import annotations

import math
from dataclasses import dataclass
import enum
from typing import Any, Dict, List, Sequence, TypeVar

from .._aliases import (
    PydanticAIAgentResult,
)

__all__ = (
    "ConfidenceLevel",
    "FieldConfidence",
    "Confidence",
    "score_confidence",
)


Output = TypeVar("Output")


_HIGH_THRESHOLD = 0.90
_MEDIUM_THRESHOLD = 0.75
_LOW_THRESHOLD = 0.50


class ConfidenceLevel(enum.Enum):
    """
    Confidence level of a semantic operation's result.
    """

    HIGH = "high"
    MEDIUM = "medium"
    LOW = "low"
    VERY_LOW = "very_low"


@dataclass
class FieldConfidence:
    """Per-field confidence breakdown for structured outputs."""

    field_name: str
    """Name of the output field."""

    confidence: float
    """Probability-based confidence score in ``[0, 1]``."""

    level: ConfidenceLevel
    """Qualitative confidence level."""

    token_count: int = 0
    """Number of tokens that were mapped to this field."""

    avg_logprob: float = 0.0
    """Mean log-probability across the tokens for this field."""


@dataclass
class Confidence:
    """Aggregated confidence information extracted from token log-probabilities.

    This is attached to :class:`ParsedResult` when the ``confidence`` flag
    is passed to ``parse`` / ``aparse``.
    """

    overall: float
    """Overall confidence score — geometric mean of per-token probabilities."""

    level: ConfidenceLevel
    """Qualitative confidence band for the overall score."""

    fields: Dict[str, FieldConfidence]
    """Per-field confidence breakdown (empty for simple/primitive outputs)."""

    token_count: int
    """Total number of content tokens that had log-probability data."""

    model: str = ""
    """Model name that produced the response."""

    @property
    def is_reliable(self) -> bool:
        """``True`` when overall confidence is :attr:`ConfidenceLevel.HIGH`."""
        return self.level is ConfidenceLevel.HIGH

    @property
    def low_confidence_fields(self) -> List[str]:
        """Field names whose confidence is LOW or VERY_LOW."""
        return [
            name
            for name, fc in self.fields.items()
            if fc.level in (ConfidenceLevel.LOW, ConfidenceLevel.VERY_LOW)
        ]

    def _should_hide_fields_in_repr(self) -> bool:
        """Hide `fields` in repr for primitive/single-value outputs.

        In primitive output cases we currently score a synthetic `"value"` field
        so we can still compute a sensible `overall` score without brittle
        string matching. That extra wrapper is an implementation detail and
        shouldn't clutter the default repr.
        """
        if not self.fields:
            return True
        if len(self.fields) != 1:
            return False
        fc = self.fields.get("value")
        return fc is not None and fc.field_name == "value"

    def __repr__(self) -> str:
        base = (
            "Confidence("
            f"overall={self.overall}, "
            f"level={self.level!r}, "
            f"token_count={self.token_count}"
        )
        if self.model:
            base += f", model={self.model!r}"

        if self._should_hide_fields_in_repr():
            return base + ")"

        return base + f", fields={self.fields!r})"

    def __rich__(self):
        from rich.console import RenderableType, Group
        from rich.rule import Rule
        from rich.text import Text

        renderables: list[RenderableType] = []

        renderables.append(
            Rule(
                title="✨ Confidence",
                style="rule.line",
                align="left",
            )
        )

        renderables.append(
            Text.from_markup(f"[bold]overall: {self.overall}[/bold]")
        )
        renderables.append(
            Text.from_markup(
                f"[dim italic]level: {self.level.value}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[dim italic]token_count: {self.token_count}[/dim italic]"
            )
        )

        if self.model:
            renderables.append(
                Text.from_markup(
                    f"[sandy_brown]>>>[/sandy_brown] [dim italic]Model: {self.model}[/dim italic]"
                )
            )

        if not self._should_hide_fields_in_repr():
            renderables.append(
                Text.from_markup(
                    f"[dim italic]fields: {self.fields}[/dim italic]"
                )
            )

        return Group(*renderables)


def _score_confidence_level(score: float) -> ConfidenceLevel:
    if score >= _HIGH_THRESHOLD:
        return ConfidenceLevel.HIGH
    if score >= _MEDIUM_THRESHOLD:
        return ConfidenceLevel.MEDIUM
    if score >= _LOW_THRESHOLD:
        return ConfidenceLevel.LOW
    return ConfidenceLevel.VERY_LOW


def _logprob_to_prob(lp: float) -> float:
    try:
        return float(math.exp(lp))
    except (OverflowError, ValueError):
        return 0.0


def _geometric_mean_prob(probs: Sequence[float]) -> float:
    """Geometric mean via log-sum, clamping near-zero values."""
    if not probs:
        return 0.0
    log_sum = sum(math.log(max(p, 1e-10)) for p in probs)
    return math.exp(log_sum / len(probs))


def _extract_logprob_tokens(
    runs: Sequence[PydanticAIAgentResult[Any]],
) -> List[Dict[str, Any]]:
    """Walk backwards through run messages and return the first list of
    token-level log-prob entries found.

    Checks both ``ModelResponse.provider_details`` (Chat API) and
    ``TextPart.provider_details`` (Responses API).
    """
    from pydantic_ai.messages import ModelResponse, TextPart

    for run in reversed(runs):
        try:
            messages = list(run.all_messages())
        except Exception:
            continue

        for msg in reversed(messages):
            if not isinstance(msg, ModelResponse):
                continue

            # --- Chat API path: logprobs on the response itself ---
            details = msg.provider_details or {}
            if isinstance(details, dict):
                logprobs = details.get("logprobs") or details.get("log_probs")
                if isinstance(logprobs, list) and logprobs:
                    return [
                        {
                            "token": item.get("token", ""),
                            "logprob": item.get("logprob"),
                            "probability": _logprob_to_prob(item["logprob"])
                            if item.get("logprob") is not None
                            else 0.0,
                        }
                        for item in logprobs
                        if isinstance(item, dict)
                    ]

            # --- Responses API path: logprobs on TextPart ---
            for part in msg.parts:
                if isinstance(part, TextPart) and part.provider_details:
                    logprobs = part.provider_details.get("logprobs")
                    if isinstance(logprobs, list) and logprobs:
                        return [
                            {
                                "token": item.get("token", ""),
                                "logprob": item.get("logprob"),
                                "probability": _logprob_to_prob(
                                    item["logprob"]
                                )
                                if item.get("logprob") is not None
                                else 0.0,
                            }
                            for item in logprobs
                            if isinstance(item, dict)
                        ]
    return []


def _map_tokens_to_fields(
    tokens: List[Dict[str, Any]],
    extracted_data: Dict[str, Any],
) -> Dict[str, List[Dict[str, Any]]]:
    """Best-effort mapping of token spans to output field values.

    Looks for the string representation of each field value within the
    concatenated token text to find the tokens that correspond to it.
    """
    field_tokens: Dict[str, List[Dict[str, Any]]] = {
        k: [] for k in extracted_data
    }
    if not tokens:
        return field_tokens

    full_text = "".join(t.get("token", "") for t in tokens)

    for field_name, value in extracted_data.items():
        if value is None:
            continue
        value_str = str(value).strip()
        if not value_str:
            continue

        start_idx = full_text.lower().find(value_str.lower())
        if start_idx == -1 and isinstance(value, (int, float)):
            start_idx = full_text.lower().find(str(value))

        if start_idx != -1:
            char_count = 0
            value_end = start_idx + len(value_str)
            for token_data in tokens:
                token = token_data.get("token", "")
                token_start = char_count
                token_end = char_count + len(token)
                if token_start < value_end and token_end > start_idx:
                    field_tokens[field_name].append(token_data)
                char_count = token_end

        # Fallback: if we couldn't locate the value, use the global average
        if not field_tokens[field_name]:
            avg_prob = (
                sum(t.get("probability", 0.0) for t in tokens) / len(tokens)
                if tokens
                else 0.5
            )
            field_tokens[field_name] = [
                {"probability": avg_prob, "token": "", "logprob": 0.0}
            ]

    return field_tokens


def _score_field(
    field_name: str,
    value: Any,
    tokens: List[Dict[str, Any]],
) -> FieldConfidence:
    if not tokens:
        return FieldConfidence(
            field_name=field_name,
            confidence=0.5,
            level=ConfidenceLevel.LOW,
        )

    probs = [t.get("probability", 0.0) for t in tokens]
    logprobs = [t.get("logprob", 0.0) for t in tokens]
    confidence = _geometric_mean_prob(probs)
    avg_lp = sum(logprobs) / len(logprobs) if logprobs else 0.0

    return FieldConfidence(
        field_name=field_name,
        confidence=round(confidence, 4),
        level=_score_confidence_level(confidence),
        token_count=len(tokens),
        avg_logprob=round(avg_lp, 4),
    )


def _is_primitive_output(output: Any) -> bool:
    """True if output is a primitive (str, int, float, bool, None)."""
    return output is None or isinstance(output, (str, int, float, bool))


def score_confidence(
    runs: Sequence[PydanticAIAgentResult[Any]],
    output: Any,
    *,
    model_name: str | None = None,
) -> Confidence | None:
    """Compute a :class:`Confidence` score from token log-probabilities.

    Handles both structured outputs (per-field mapping) and primitive outputs
    (str, int, float, bool): for primitives the entire token sequence is used
    to compute a single overall confidence so scoring does not depend on
    string-matching the value in the token stream.

    Returns ``None`` when no usable log-prob data is available (e.g. the
    model/provider does not support log-probabilities).
    """
    tokens = _extract_logprob_tokens(runs)
    if not tokens:
        return None

    # Build a dict of fields for structured outputs; primitives get a single "value" key
    if hasattr(output, "model_dump"):
        output_map: Dict[str, Any] = output.model_dump()
    elif isinstance(output, dict):
        output_map = output
    else:
        output_map = {"value": output}

    is_primitive = _is_primitive_output(output)

    if is_primitive:
        # For primitives, use the full token sequence for the single "value" field
        # so we don't rely on string-matching (e.g. JSON "true" vs True, or spacing).
        field_token_map = {"value": tokens}
    else:
        field_token_map = _map_tokens_to_fields(tokens, output_map)

    field_results: Dict[str, FieldConfidence] = {}
    for fname, fvalue in output_map.items():
        field_results[fname] = _score_field(
            fname, fvalue, field_token_map.get(fname, [])
        )

    if field_results:
        overall = sum(fc.confidence for fc in field_results.values()) / len(
            field_results
        )
    else:
        probs = [t.get("probability", 0.0) for t in tokens]
        overall = _geometric_mean_prob(probs) if probs else 0.0

    return Confidence(
        overall=round(overall, 4),
        level=_score_confidence_level(overall),
        fields=field_results,
        token_count=len(tokens),
        model=model_name or "",
    )



"""zyx._graph._requests"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Generic, Type, TypeVar

from pydantic_ai.output import NativeOutput

from .._processing._toon import object_as_text
from .._processing._messages import (
    parse_instructions_as_system_prompt_parts,
)
from .._aliases import (
    PydanticAIModelRequest,
    PydanticAIMessage,
    PydanticAIToolset,
    PydanticAISystemPromptPart,
)
from .._utils._outputs import OutputBuilder
from .._utils._strategies._params import SourceStrategy, TargetStrategy
from ..attachments import Attachment, is_attachment_like
from ._ctx import (
    SemanticGraphContext,
)

__all__ = ("SemanticGraphRequestTemplate",)


Deps = TypeVar("Deps")
Output = TypeVar("Output")


@dataclass
class SemanticGraphRequestTemplate(Generic[Output]):
    """
    Template class used by various steps within a semantic operation's graph, to build
    the context of a single request.
    """

    output_type: Type[Output] | None = None
    """An override to the `output_type` parameter within a single step/agent run within a
    graph."""

    native_output: bool = False
    """Whether to use `pydantic_ai`'s `NativeOutput` type for the output of this request,
    this allows the model to use it's native structured output capabilities."""

    native_output_name: str | None = None
    """The name of the output to be used for this request."""

    native_output_description: str | None = None
    """The description of the output to be used for this request."""

    system_prompt_additions: str | list[str] | None = None
    """Dynamic instructions that can be applied onto the existing system prompt/instructions
    within `SemanticGraphDeps`. These are added before the primary input (source) block."""

    user_prompt_additions: str | list[str] | None = None
    """Dynamic user prompt content to append to the message history of a request."""

    toolsets: List[PydanticAIToolset] | None = None
    """Any operation-specific toolsets that can be included within the request."""

    include_run_context: bool = True
    """Whether to include the context of previous runs within the request."""

    include_source_context: bool = True
    """Whether to include context about a given `source` within the system prompt of a
    request. (If one is provided)"""

    include_output_context: bool = False
    """Whether to provide context about the current state of the output within the system
    prompt of a request."""

    include_user_toolsets: bool = True
    """Whether user-provided tools/toolsets should be included within this request."""

    def render(
        self, ctx: SemanticGraphContext[Output, Deps]
    ) -> Dict[str, Any]:
        """
        Renders a dictionary of parameters compatible for invoking a `pydantic_ai.Agent` instance
        using the deps & state within a `SemanticGraphContext` object.
        """
        message_history = self._build_message_history(ctx)
        system_parts = self._build_system_prompt(ctx, message_history)
        if system_parts:
            message_history = [
                PydanticAIModelRequest(parts=system_parts),
                *message_history,
            ]

        self._apply_user_prompt_additions(message_history)
        toolsets = self._collect_toolsets(ctx)
        output_type = self._resolve_output_type(ctx)

        return {
            "message_history": message_history,
            "output_type": output_type,
            "toolsets": toolsets,
            "deps": ctx.deps.deps,
            "usage_limits": getattr(ctx.deps, "usage_limits", None),
        }

    def _build_message_history(
        self, ctx: SemanticGraphContext[Output, Deps]
    ) -> List[PydanticAIMessage]:
        message_history: List[PydanticAIMessage] = (
            list(ctx.deps.message_history) if ctx.deps.message_history else []
        )

        if self.include_run_context and hasattr(ctx.state, "agent_runs"):
            for run in ctx.state.agent_runs:
                message_history.extend(run.new_messages())

        return message_history

    def _build_system_prompt(
        self,
        ctx: SemanticGraphContext[Output, Deps],
        message_history: List[PydanticAIMessage],
    ) -> List[PydanticAISystemPromptPart]:
        system_parts: List[PydanticAISystemPromptPart] = (
            list(ctx.deps.instructions) if ctx.deps.instructions else []
        )

        system_parts.extend(self._add_target_instructions(ctx))
        system_parts.extend(self._add_system_prompt_additions(ctx))
        system_parts.extend(self._add_attachments(ctx, message_history))

        if self.include_source_context:
            system_parts.extend(self._add_source_context(ctx, message_history))

        if (
            self.include_output_context
            and getattr(ctx.state, "output", None) is not None
        ):
            system_parts.append(
                PydanticAISystemPromptPart(
                    content=_render_output_context(ctx.state.output)
                )
            )

        return system_parts

    def _add_target_instructions(
        self, ctx: SemanticGraphContext[Output, Deps]
    ) -> List[PydanticAISystemPromptPart]:
        target_strategy: TargetStrategy | None = getattr(
            ctx.deps, "target_strategy", None
        )
        if target_strategy is None:
            return []
        if not target_strategy.instructions:
            return []
        return parse_instructions_as_system_prompt_parts(
            instructions=target_strategy.instructions,
            deps=ctx.deps.deps,
        )

    def _add_system_prompt_additions(
        self, ctx: SemanticGraphContext[Output, Deps]
    ) -> List[PydanticAISystemPromptPart]:
        if not self.system_prompt_additions:
            return []
        return parse_instructions_as_system_prompt_parts(
            instructions=self.system_prompt_additions,  # type: ignore
            deps=ctx.deps.deps,
        )

    def _add_attachments(
        self,
        ctx: SemanticGraphContext[Output, Deps],
        message_history: List[PydanticAIMessage],
    ) -> List[PydanticAISystemPromptPart]:
        parts: List[PydanticAISystemPromptPart] = []
        if not getattr(ctx.deps, "attachments", None):
            return parts

        for attachment in ctx.deps.attachments:  # type: ignore
            if isinstance(attachment, Attachment):
                parts.append(
                    PydanticAISystemPromptPart(
                        content=_render_attachment_context(
                            name=attachment.name
                            or type(attachment.source).__name__,
                            description=attachment.get_description(),
                            state=attachment.get_state_description(),
                        )
                    )
                )
                if attachment.message is not None:
                    if attachment.message not in message_history:
                        message_history.insert(0, attachment.message)
                continue
            if is_attachment_like(attachment):
                name = getattr(attachment, "name", None)
                parts.append(
                    PydanticAISystemPromptPart(
                        content=_render_attachment_context(
                            name=name or type(attachment).__name__,
                            description=attachment.get_description(),
                            state=attachment.get_state_description(),
                        )
                    )
                )
                message = getattr(attachment, "message", None)
                if message is not None:
                    if message not in message_history:
                        message_history.insert(0, message)
                continue

            parts.append(
                PydanticAISystemPromptPart(
                    content=_render_attachment_context(
                        name="Attachment",
                        description=str(attachment),
                        state=None,
                    )
                )
            )

        return parts

    def _add_source_context(
        self,
        ctx: SemanticGraphContext[Output, Deps],
        message_history: List[PydanticAIMessage],
    ) -> List[PydanticAISystemPromptPart]:
        parts: List[PydanticAISystemPromptPart] = []
        source_strategy: SourceStrategy | None = getattr(
            ctx.deps, "source_strategy", None
        )
        if source_strategy is None:
            if ctx.deps.source is None:
                return parts
            parts.append(
                PydanticAISystemPromptPart(
                    content=_render_source_context(ctx.deps.source)
                )
            )
            return parts

        payload = source_strategy.get_payload()
        if payload.kind == "raw":
            parts.append(
                PydanticAISystemPromptPart(
                    content=_render_source_context(payload.content)
                )
            )
            return parts

        parts.append(
            PydanticAISystemPromptPart(
                content=_render_source_metadata_values(
                    origin=payload.origin,
                    media_type=payload.media_type,
                    source_repr=payload.source_repr,
                )
            )
        )
        parts.append(
            PydanticAISystemPromptPart(
                content=_render_source_attachment_context(payload.content)
            )
        )
        if payload.message is not None:
            if payload.message not in message_history:
                message_history.insert(0, payload.message)
        return parts

    def _apply_user_prompt_additions(
        self, message_history: List[PydanticAIMessage]
    ) -> None:
        if not self.user_prompt_additions:
            return
        message_history.append(
            PydanticAIModelRequest.user_text_prompt(
                user_prompt=str(self.user_prompt_additions)
            )
        )

    def _collect_toolsets(
        self, ctx: SemanticGraphContext[Output, Deps]
    ) -> List[PydanticAIToolset]:
        toolsets: List[PydanticAIToolset] = []
        if self.include_user_toolsets and ctx.deps.toolsets:
            toolsets.extend(ctx.deps.toolsets)
        if self.toolsets:
            toolsets.extend(self.toolsets)
        return toolsets

    def _resolve_output_type(
        self, ctx: SemanticGraphContext[Output, Deps]
    ) -> Type[Output] | None:
        target_strategy: TargetStrategy | None = getattr(
            ctx.deps, "target_strategy", None
        )
        if target_strategy is None:
            return self.output_type

        if not target_strategy.is_target_wrapper:
            output_type = self.output_type
            if (
                output_type is None
                and getattr(ctx.state, "output", None) is not None
            ):
                output_type = getattr(ctx.state.output, "normalized", None)

            if self.native_output:
                if output_type is not str:
                    output_type = NativeOutput(
                        outputs=output_type,
                        name=target_strategy.name or self.native_output_name,
                        description=target_strategy.description
                        or self.native_output_description,
                    )
            return output_type  # type: ignore

        output_type = NativeOutput(
            outputs=ctx.state.output.normalized,
            name=target_strategy.name,
            description=target_strategy.description,
        )
        if target_strategy.has_hooks:
            return None

        return output_type  # type: ignore


def _render_output_context(builder: OutputBuilder[Output]) -> str:
    """Generates system prompt context for the current state of the output
    builder.
    """
    if builder.is_value:
        context = (
            f"\n\n[Output Context]\n"
            f"You are currently building/generating an output of type: {object_as_text(builder.normalized)}\n"
            f"The output had a starting state of: {object_as_text(builder.target)}\n"
        )
    else:
        context = (
            f"\n\n[Output Context]\n"
            f"You are currently building/generating an output of type: {object_as_text(builder.normalized)}\n"
        )

    if builder.partial is not None:
        context += f"The current state of the output is: {object_as_text(builder.partial)}\n"

    return context


def _render_source_context(source: Any | Type[Any]) -> str:
    """Generates system prompt context for a source object or value.

    The only content between the markers is the raw source; no instructional
    line inside the block, so the model returns exactly that content when
    asked to extract the primary input.
    """
    body = object_as_text(source)
    return f"\n\n[PRIMARY INPUT]\n\n{body}\n\n[END PRIMARY INPUT]\n"


def _render_source_attachment_context(content: str) -> str:
    """Generates system prompt context for a source attachment."""
    return f"\n\n[PRIMARY INPUT]\n\n{content}\n\n[END PRIMARY INPUT]\n"


def _render_source_metadata_values(
    *, origin: Any, media_type: Any, source_repr: str | None
) -> str:
    """Generates system prompt metadata for a source."""
    origin_value = getattr(origin, "value", origin)
    media_value = getattr(media_type, "value", media_type)
    source_value = source_repr or ""
    return (
        "\n\n[SOURCE META]\n"
        f"Origin: {origin_value}\n"
        f"Media Type: {media_value}\n"
        f"Source: {source_value}\n"
        "[END SOURCE META]\n"
    )


def _render_attachment_context(
    name: str | None,
    description: str,
    state: str | None,
) -> str:
    title = name or "Attachment"
    body = f"\n\n[ATTACHMENT: {title}]\n"
    body += f"Description:\n{description}\n"
    if state:
        body += f"\nState:\n{state}\n"
    body += "[END ATTACHMENT]\n"
    return body



"""zyx.result"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import (
    Any,
    Generic,
    List,
    TypeVar,
)

from ._aliases import (
    PydanticAIAgentResult,
    PydanticAIUsage,
    PydanticAIMessage,
)
from ._utils._confidence import Confidence, score_confidence
from ._utils._semantic import semantic_from_output, semantic_for_operation

__all__ = ("Result",)


Output = TypeVar("Output")


@dataclass
class Result(Generic[Output]):
    """
    Universal representation of the output/result of a semantic operation.

    This class is used to wrap the final `output`, which provides either a
    complete or partial representation of the `target` type or value, based on
    the operation used.
    """

    output: Output
    """
    The final output content generated by the semantic operation.

    This is either a complete, partial or mutated representation of the `target`
    type or value."""

    raw: List[PydanticAIAgentResult[Any]] = field(repr=False)
    """All underlying AgentRunResult(s) that were used to produce this result."""

    _semantic_message: str | None = field(default=None, init=False)

    @property
    def confidence(self) -> Confidence:
        """
        Confidence scores for the `output` of this result.

        NOTE: This may return `None` if `confidence` was not enabled when running the
        semantic operation this result was produced from, or if the model API does not support
        retrieving log-probabilities.

        Returns:
            A `Confidence` object containing the confidence scores for the `output` of this result.
        """
        return score_confidence(
            runs=self.raw,
            output=self.output,
            model_name=self.model,
        )  # type: ignore

    @property
    def model(self) -> str:
        """
        The model that was used to produce this result.
        """
        return self.raw[0].response.model_name or ""

    @property
    def usage(self) -> PydanticAIUsage:
        """
        The total accumulated token usage across all agent runs that were used internally
        to produce this result.
        """
        total = PydanticAIUsage()
        for run in self.raw:
            total.incr(run.usage())
        return total

    @property
    def all_new_messages(self) -> List[PydanticAIMessage]:
        """
        All new messages genenerated by `pydantic_ai` that were added/utilized during the
        execution of the operation that was used to produce this result.
        """
        messages = []

        for run in self.raw:
            messages.extend(run.new_messages())

        return messages

    def __repr__(self) -> str:
        return (
            f"Result({type(self.output).__name__}):\n"
            f"{self.output}\n\n"
            f">>> Model: {self.model}\n"
        )

    def as_message(
        self,
        *,
        operation: str | None = None,
        original: Any | None = None,
        updated: Any | None = None,
        summary: str | None = None,
    ) -> str:
        """Return a semantic message for this result."""
        if self._semantic_message:
            return self._semantic_message
        if operation:
            return semantic_for_operation(
                operation,
                output=self.output,
                original=original,
                updated=updated,
                summary=summary,
            )
        return semantic_from_output(self.output)

    def __rich__(self):
        from rich.console import RenderableType, Group
        from rich.rule import Rule
        from rich.text import Text

        renderables: list[RenderableType] = []

        renderables.append(
            Rule(
                title=f"✨ Result({type(self.output).__name__})",
                style="rule.line",
                align="left",
            )
        )

        output_text = f"{self.output}\n"
        renderables.append(Text.from_markup(f"[bold]{output_text}[/bold]"))

        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Type: {type(self.output).__name__}[/dim italic]"
            )
        )

        model = getattr(self, "model", None)
        if model:
            model_names = (
                model
                if isinstance(model, str)
                else (model[0] if len(model) == 1 else ", ".join(model))
            )
            label = (
                "Model"
                if (
                    isinstance(model_names, str)
                    or (hasattr(model, "__len__") and len(model) == 1)
                )
                else "Models"
            )
            model_info = Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]{label}: {model_names}[/dim italic]"
            )
            renderables.append(model_info)

        return Group(*renderables)



from __future__ import annotations

import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Literal, TypeAlias

try:
    import chromadb  # type: ignore[unresolved-import]
except ImportError:
    raise ImportError(
        "To use the `chroma` memory provider, you must first install the `chromadb` library.\n"
        "You can install it by using one of the following commands:\n"
        "```bash\n"
        "pip install zyx[chroma]\n"
        "pip install chromadb\n"
        "```"
    )

from . import MemoryProvider

__all__ = (
    "ChromaMemoryProvider",
    "ChromaMemoryClientKind",
)


ChromaMemoryClientKind: TypeAlias = Literal["persistent", "ephemeral"]


@dataclass(init=False)
class ChromaMemoryProvider(MemoryProvider):
    """
    Memory provider that utilizes `chromadb`.
    """

    @property
    def client(self) -> chromadb.ClientAPI | None:
        """The `chromadb.ClientAPI` instance to use for this memory provider."""
        return getattr(self, "_client", None)

    @property
    def path(self) -> str | Path | None:
        """An explicit directory path to use for the persistent client, if not
        provided this will use the current working directory + `memory/chroma`.
        """
        return getattr(self, "_path", None)

    @property
    def collection_name(self) -> str:
        """An optional name for the collection to use for this memory provider.

        This can contain the `{key}` placeholder, which will inject the name
        of the `Memory` resource using this provider."""
        return getattr(self, "_collection_name", "memory-{key}")

    @collection_name.setter
    def collection_name(self, value: str) -> None:
        self._collection_name = value

    def __init__(
        self,
        client: chromadb.ClientAPI | ChromaMemoryClientKind = "persistent",
        path: str | Path | None = None,
        collection_name: str = "memory-{key}",
    ):
        """
        Initialize a new `ChromaMemoryProvider` instance.

        Args:
            client : `chromadb.ClientAPI` | `ChromaMemoryClient`
                The `chromadb.ClientAPI` instance to use for this memory provider, or the type of client to use.
            path : `str` | `Path` | None
                An explicit directory path to use for the persistent client, if not
                provided this will use the current working directory + `memory/chroma`.
            collection_name : `str`
                An optional name for the collection to use for this memory provider.

                This can contain the `{key}` placeholder, which will inject the name
                of the `Memory` resource using this provider.
        """
        if client == "persistent":
            if not self.path:
                path = Path.cwd() / "memory/chroma"
            self._client = chromadb.PersistentClient(path=str(path))
        else:
            self._client = chromadb.EphemeralClient()

        if path:
            self._path = path

        self._collection_name = collection_name

    def get_collection(self, key: str) -> chromadb.Collection:
        if self.client is not None and hasattr(
            self.client, "get_or_create_collection"
        ):
            return self.client.get_or_create_collection(
                self.collection_name.format(key=key)
            )

    async def add(self, key: str, content: str) -> str:
        collection = self.get_collection(key)
        id = str(uuid.uuid4())

        collection.add(documents=[content], metadatas=[{"id": id}], ids=[id])
        return id

    async def delete(self, key: str, id: str) -> None:
        collection = self.get_collection(key)
        collection.delete(ids=[id])

    async def search(
        self, key: str, query: str, n: int = 20
    ) -> Dict[str, str]:
        collection = self.get_collection(key)

        results = collection.query(query_texts=[query], n_results=n)
        if not results.get("ids") or not results.get("documents"):
            return {}

        return dict(
            zip(
                results["ids"][0],
                results["documents"][0] if results["documents"] else [],
                strict=False,
            )
        )



"""zyx._types

Framework specific types and aliases most commonly found within the function signatures
of semantic operations.
"""

from __future__ import annotations

from typing import (
    Any,
    Callable,
    Dict,
    Literal,
    Type,
    TypeAlias,
    TypeVar,
    Union,
)

from pydantic import BaseModel
from pydantic_ai import (
    agent as _pydantic_ai_agent,
    models as _pydantic_ai_models,
)

from ._aliases import (
    PydanticAIAgent,
    PydanticAIModel,
    PydanticAIMessage,
    PydanticAITool,
    PydanticAIToolset,
    PydanticAIBuiltinTool,
)
from .attachments import Attachment, AttachmentLike
from .context import Context
from .tools.memory import Memory
from .tools.code import Code

__all__ = (
    "Deps",
    "Output",
    "InstructionsParam",
    "SourceParam",
    "TargetParam",
    "ModelParam",
    "ContextType",
    "ToolType",
    "AttachmentType",
)


Deps = TypeVar("Deps")
"""Alias for `deps` used within `pydantic_ai.RunContext`."""


Output = TypeVar("Output")
"""Generic type variable for the final output found within the `Result` or
`Stream` returned by a semantic operation."""


InstructionsParam: TypeAlias = _pydantic_ai_agent.Instructions | Any
"""Accepted formats of passing in instructions to a semantic operation. This is a re-export
of the `pydantic_ai.agent.Instructions` type, and supports:

- Strings / Lists of strings
- Functions that return a string / and can optionally accept `RunContext` dependencies
"""


SourceParam: TypeAlias = Any
"""The input object to be considered the source of truth or 'main' data/context within
a semantic operation."""


TargetParam: TypeAlias = Union[Output, Type[Output]]
"""The output `target` - a type, instance, or value to generate/modify through
a semantic operation."""


KnownModelName: TypeAlias = Literal[
    "ollama:functiongemma",
    "ollama:deepseek-ocr",
    "ollama:deepseek-v3.1",
    "ollama:devstral-small-2",
    "ollama:devstral-small-2:cloud",
    "ollama:lfm2.5-thinking",
    "ollama:glm-4.7-flash",
    "ollama:glm-5:cloud",
    "ollama:glm-ocr",
    "ollama:gpt-oss",
    "ollama:gpt-oss-safeguard",
    "ollama:granite4",
    "ollama:kimi-k2.5:cloud",
    "ollama:minimax-m2.5:cloud",
    "ollama:ministral-3",
    "ollama:nemotron-3-nano",
    "ollama:olmo-3",
    "ollama:olmo-3.1",
    "ollama:translategemma",
    "ollama:rnj-1",
    "ollama:qwen3-coder-next",
    "ollama:qwen3-coder-next:cloud",
    "ollama:qwen3-next",
    "ollama:qwen3-vl",
]


ModelParam: TypeAlias = Union[
    PydanticAIAgent,
    PydanticAIModel,
    KnownModelName,
    _pydantic_ai_models.KnownModelName,
    str,
]
"""Accepted formats for a `runner` or generator that will be used to produce the output
of a semantic operation.

This can be passed as:
- A string name of a known model in the `pydantic_ai` format (e.g. "openai:gpt-4o-mini")
- A PydanticAI Model (OpenAIChatModel, AnthropicModel, etc.)
- A PydanticAI Agent
"""


ContextType: TypeAlias = Union[
    str,
    PydanticAIMessage,
    BaseModel,
    Dict[str, Any],
    Context,
    AttachmentLike,
    Attachment
]
"""Accepted formats for a single item within the ``context`` of a semantic
operation.  All items are converted to ``pydantic_ai.ModelMessage`` objects
before invocation.

Accepted types:

- A :class:`~zyx.context.Context` accumulates conversation state across
  operations.  When passed, its messages and rendered instructions are
  injected into the request, and it is auto-updated after the operation
  when ``Context.update`` is ``True``.

- A string (with optional role tags like ``[system]``/``[s]``,
  ``[user]``/``[u]``, ``[assistant]``/``[a]``)

- A dictionary or Pydantic model in the PydanticAI or OpenAI Chat Completions
  format

- A ``pydantic_ai.ModelMessage``

Examples::

    ctx = Context()
    ctx.add_user("Hi!")
    result = zyx.make(context=[ctx, "What did I just say?"], target=str)
"""


ToolType: TypeAlias = Union[
    Callable[..., Any],
    PydanticAITool,
    PydanticAIBuiltinTool,
    PydanticAIToolset,
    Memory,
    Code,
]
"""
Accepted formats in which a single item within the `tools` parameter of a semantic operation
can be passed in as. All tools support passing in RunContext dependencies.

This is a unification of all representations of a tool that are provided by `pydantic_ai`,
including:

- Functions
- PydanticAI Tools
- PydanticAI Builtin Tools
- PydanticAI Toolsets
"""


AttachmentType: TypeAlias = Union[Any, Type[Any], Attachment, AttachmentLike]
"""
Accepted formats in which a single item within the `attachments` parameter of a semantic operation
can be passed in as.
"""



"""zyx._utils._strategies._abstract"""

from __future__ import annotations

from abc import ABC, abstractmethod

from ..._aliases import PydanticAIToolset

__all__ = ("AbstractStrategy",)


class AbstractStrategy(ABC):
    """
    Base class for 'strategies' that are provided to agents or models
    to interface with various objects and resources.
    """

    @property
    @abstractmethod
    def kind(self) -> str:
        raise NotImplementedError

    def get_description(self) -> str:
        return self.kind

    def get_state_description(self) -> str | None:
        return None

    def get_toolset(self) -> PydanticAIToolset | None:
        return None



"""zyx

> A fun **"anti-framework"** for doing useful things with agents and LLMs.

## Quick Start

### Generating Content

```python
from zyx import make

def get_weather(city: str) -> str:
    return f"The weather in {city} is sunny."

result = make(
    target=str,
    context="What is the weather in Tokyo?",
    tools=[get_weather],
)
```

### Streaming

```python
from zyx import make

stream = make(context="Write a haiku about the weather.", stream=True)

for chunk in stream.text():
    print(chunk, end="", flush=True)
```

---

### Editing Values

```python
from zyx import edit

data = {
    "name": "John Doe",
    "age": 30,
    "email": "john.doe@example.com",
}

result = edit(
    target=data,
    context="Change the name to 'Jane Doe'.",
    merge=False,
)
```
"""

# --- core components & objects ---
from .attachments import Attachment, paste, attach
from .context import Context, create_context
from .targets import Target, target
from .result import Result
from .stream import Stream

# --- tools / prebuilt toolsets ---
from .tools import Tool, tool, Code, Memory, memories

# --- semantic operations ---
from .operations.edit import aedit, edit
from .operations.expressions import expr
from .operations.make import amake, make
from .operations.parse import aparse, parse
from .operations.query import aquery, query
from .operations.run import arun, run
from .operations.select import aselect, select
from .operations.validate import avalidate, validate

__all__ = (
    # --- core components & objects ---
    "Attachment",
    "paste",
    "attach",
    "Context",
    "create_context",
    "Target",
    "target",
    "Result",
    "Stream",
    # --- tools / prebuilt toolsets ---
    "Tool",
    "tool",
    "Code",
    "Memory",
    "memories",
    # --- semantic operations ---
    "aedit",
    "edit",
    "expr",
    "amake",
    "make",
    "aparse",
    "parse",
    "aquery",
    "query",
    "arun",
    "run",
    "aselect",
    "select",
    "avalidate",
    "validate",
)



"""zyx.tools

Provides pre-built tools/toolsets that can be used by a model/agent
for various purposes, such as memory and code execution.

Various types within this module can be added within the `attachments`
parameter of a semantic operation, for further functionality.
"""

from .._aliases import PydanticAITool
from .tool import Tool, tool
from .code import Code
from .memory import Memory, memories

__all__ = (
    "PydanticAITool",
    "Tool",
    "tool",
    "Code",
    "Memory",
    "memories",
)



"""zyx.operations.run"""

from __future__ import annotations

import asyncio
import inspect
from dataclasses import dataclass, field
from types import SimpleNamespace
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    Optional,
    TypeVar,
    cast,
    overload,
    TYPE_CHECKING,
)

from pydantic import BaseModel
import pydantic_monty
from pydantic_ai.exceptions import ModelRetry
from pydantic_ai.tools import RunContext as PydanticAIRunContext
from pydantic_ai.toolsets import FunctionToolset

from .._aliases import (
    PydanticAIInstructions,
    PydanticAIModelSettings,
    PydanticAIUsage,
    PydanticAIUsageLimits,
    PydanticAIModel,
)
from .._graph import SemanticGraphDeps, SemanticGraphRequestTemplate
from .._processing._toon import object_as_text
from .._types import (
    AttachmentType,
    ContextType,
    Deps,
    ModelParam,
    SourceParam,
    TargetParam,
    ToolType,
)
from .._utils._outputs import OutputBuilder
from .._utils._semantic import semantic_for_operation
from ..result import Result
from .make import amake
from ..stream import Stream, StreamFieldMapping as StreamStreamFieldMapping
from ..targets import Target

if TYPE_CHECKING:
    from .._utils._observer import Observer

__all__ = (
    "RunCompletion",
    "RunTask",
    "RunVerificationContext",
    "RunResult",
    "RunStream",
    "arun",
    "run",
)


Output = TypeVar("Output")


_RUN_SYSTEM_PROMPT = (
    "\n[INSTRUCTION]\n"
    "- You are executing a task. Use the available tools and attachments to complete it.\n"
    "- You MUST call the `complete_task` tool exactly once when the task is complete.\n"
    "- Do not claim completion in plain text; completion happens only via the tool call.\n"
    "- If you are missing information, use tools or ask for clarification (via tools if available).\n"
    "- Your completion will be deterministically verified. If verification fails, you must try again.\n"
)

_PLAN_SYSTEM_PROMPT = (
    "\n[INSTRUCTION]\n"
    "- Produce a deterministic execution plan for this task as pure Python code.\n"
    "- Output ONLY Python code. No prose, no JSON, no markdown fences.\n"
    "- Call tools directly by name, e.g. front_door_light(False).\n"
    '- Use `prompt("...")` to queue follow-up instructions.\n'
    "- Use `complete(...)` to finalize the task.\n"
    "- Do NOT use await, multi_tool_use, or functions.* prefixes.\n"
)


class RunCompletion(BaseModel):
    """Structured completion payload emitted by the task completion tool."""

    status: Literal["success", "failed", "skipped"] = "success"
    summary: str
    result: Any | None = None
    evidence: Any | None = None
    verification: Any | None = None


@dataclass
class RunTask(Generic[Output]):
    """Lightweight task metadata for a run."""

    goal: str | None
    target: TargetParam[Output] | None
    source: SourceParam | None


@dataclass
class RunVerificationContext(Generic[Output]):
    """Context passed to a verifier for deterministic validation."""

    target: TargetParam[Output] | None
    source: SourceParam | None
    deps: Deps | None
    attachments: List[AttachmentType] | None


@dataclass
class RunResult(Result[Optional[Output]]):
    """Result wrapper for run() with completion details."""

    completion: RunCompletion | None = None
    verified: bool | None = None
    verification_error: str | None = None
    task: RunTask[Output] | None = None

    def __repr__(self) -> str:
        completion = self.completion.status if self.completion else "None"
        task_goal = self.task.goal if self.task else None
        return (
            f"RunResult({type(self.output).__name__}):\n"
            f"{self.output}\n\n"
            f">>> Model: {self.model}\n"
            f">>> Completion: {completion}\n"
            f">>> Verified: {self.verified}\n"
            f">>> Task: {task_goal}\n"
        )

    def __rich__(self):
        from rich.console import RenderableType, Group
        from rich.rule import Rule
        from rich.text import Text

        renderables: list[RenderableType] = []

        renderables.append(
            Rule(
                title=f"✨ RunResult({type(self.output).__name__})",
                style="rule.line",
                align="left",
            )
        )

        output_text = f"{self.output}\n"
        renderables.append(Text.from_markup(f"[bold]{output_text}[/bold]"))

        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Type: {type(self.output).__name__}[/dim italic]"
            )
        )

        model = getattr(self, "model", None)
        if model:
            model_names = (
                model
                if isinstance(model, str)
                else (model[0] if len(model) == 1 else ", ".join(model))
            )
            label = (
                "Model"
                if (
                    isinstance(model_names, str)
                    or (hasattr(model, "__len__") and len(model) == 1)
                )
                else "Models"
            )
            renderables.append(
                Text.from_markup(
                    f"[sandy_brown]>>>[/sandy_brown] [dim italic]{label}: {model_names}[/dim italic]"
                )
            )

        completion = self.completion.status if self.completion else "None"
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Completion: {completion}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Verified: {self.verified}[/dim italic]"
            )
        )
        task_goal = self.task.goal if self.task else None
        if task_goal:
            renderables.append(
                Text.from_markup(
                    f"[sandy_brown]>>>[/sandy_brown] [dim italic]Task: {task_goal}[/dim italic]"
                )
            )

        return Group(*renderables)


@dataclass
class RunStream(Stream[Output]):
    completion: RunCompletion | None = None
    verified: bool | None = None
    verification_error: str | None = None
    task: RunTask[Output] | None = None
    _require_completion: bool = True
    _execution_state: RunExecutionState[Output] | None = None

    async def finish_async(self) -> RunResult[Output]:  # type: ignore[override]
        res = await super().finish_async()
        completion_state = self.completion
        if self._execution_state is not None:
            completion_state = (
                completion_state or self._execution_state.completion_state
            )
        if self._require_completion and completion_state is None:
            raise ValueError(
                "Task did not complete. The model must call `complete`."
            )
        completion_state = completion_state or RunCompletion(
            status="skipped",
            summary="No completion tool call was made.",
        )
        return RunResult(
            output=res.output,
            raw=res.raw,
            completion=completion_state,
            verified=self.verified
            if self._execution_state is None
            else self._execution_state.verified,
            verification_error=self.verification_error
            if self._execution_state is None
            else self._execution_state.verification_error,
            task=self.task,
        )

    def finish(self) -> RunResult[Output]:  # type: ignore[override]
        res = super().finish()
        completion_state = self.completion
        if self._execution_state is not None:
            completion_state = (
                completion_state or self._execution_state.completion_state
            )
        if self._require_completion and completion_state is None:
            raise ValueError(
                "Task did not complete. The model must call `complete`."
            )
        completion_state = completion_state or RunCompletion(
            status="skipped",
            summary="No completion tool call was made.",
        )
        return RunResult(
            output=res.output,
            raw=res.raw,
            completion=completion_state,
            verified=self.verified
            if self._execution_state is None
            else self._execution_state.verified,
            verification_error=self.verification_error
            if self._execution_state is None
            else self._execution_state.verification_error,
            task=self.task,
        )

    def __repr__(self) -> str:
        completion = self.completion.status if self.completion else "None"
        task_goal = self.task.goal if self.task else None
        return (
            f"RunStream({type(self._builder.normalized).__name__}):\n"
            f" >>> Is Streaming: {self.is_streaming}\n"
            f" >>> Is Complete: {self._is_complete}\n"
            f" >>> Completion: {completion}\n"
            f" >>> Verified: {self.verified}\n"
            f" >>> Task: {task_goal}\n"
        )

    def __rich__(self):
        from rich.console import RenderableType, Group
        from rich.rule import Rule
        from rich.text import Text

        renderables: list[RenderableType] = []

        renderables.append(
            Rule(title="✨ RunStream", style="rule.line", align="left")
        )

        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Type: {type(self._builder.normalized).__name__}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Is Streaming: {self.is_streaming}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Is Complete: {self._is_complete}[/dim italic]"
            )
        )

        completion = self.completion.status if self.completion else "None"
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Completion: {completion}[/dim italic]"
            )
        )
        renderables.append(
            Text.from_markup(
                f"[sandy_brown]>>>[/sandy_brown] [dim italic]Verified: {self.verified}[/dim italic]"
            )
        )
        task_goal = self.task.goal if self.task else None
        if task_goal:
            renderables.append(
                Text.from_markup(
                    f"[sandy_brown]>>>[/sandy_brown] [dim italic]Task: {task_goal}[/dim italic]"
                )
            )

        if self._streams and len(self._streams) > 0:
            first_stream = self._streams[0]
            if (
                first_stream is not None
                and hasattr(first_stream, "response")
                and first_stream.response
                and first_stream.response.model_name
            ):
                renderables.append(
                    Text.from_markup(
                        f"[sandy_brown]>>>[/sandy_brown] [dim italic]Model: {first_stream.response.model_name}[/dim italic]"
                    )
                )

        return Group(*renderables)


@dataclass
class RunExecutionState(Generic[Output]):
    completion_state: RunCompletion | None = None
    validated_output: Output | None = None
    verified: bool | None = None
    verification_error: str | None = None
    tool_call_log: Dict[str, int] = field(default_factory=dict)
    total_tool_calls: int = 0


RunVerifier = (
    Callable[[RunCompletion], bool | str | None]
    | Callable[
        [RunCompletion, RunVerificationContext[Output]], bool | str | None
    ]
)

RunCheck = (
    Callable[[RunCompletion], bool | str | None]
    | Callable[
        [RunCompletion, RunVerificationContext[Output]], bool | str | None
    ]
)


def _build_target_schema_text(
    target: TargetParam[Output] | None,
) -> str | None:
    if target is None:
        return None

    if isinstance(target, Target):
        schema_target = target.target
        target_name = target.name
        target_desc = target.description
    else:
        schema_target = target
        target_name = None
        target_desc = None

    schema_text = object_as_text(schema_target)
    if target_name or target_desc:
        meta = []
        if target_name:
            meta.append(f"Name: {target_name}")
        if target_desc:
            meta.append(f"Description: {target_desc}")
        schema_text = "\n".join([*meta, "", schema_text])

    return schema_text


def _render_tool_inventory(toolsets: List[Any]) -> str | None:
    tools: List[str] = []
    for toolset_item in toolsets:
        if isinstance(toolset_item, FunctionToolset):
            for name, tool in toolset_item.tools.items():
                description = getattr(tool, "description", None) or ""
                if description:
                    tools.append(f"- {name}: {description}")
                else:
                    tools.append(f"- {name}")
    if not tools:
        return None
    return (
        "\n[AVAILABLE TOOLS]\n"
        + "\n".join(tools)
        + "\n[END AVAILABLE TOOLS]\n"
    )


def _observe_model_request(observe: Any | None) -> None:
    if observe and hasattr(observe, "on_model_request"):
        observe.on_model_request()


def _normalize_observer(observe: Any | None) -> Any | None:
    if not observe:
        return None
    from .._utils._observer import Observer

    if observe is True:
        return Observer()
    if isinstance(observe, Observer):
        return observe
    raise ValueError("Invalid 'observe' value. Expected bool or Observer.")


def _normalize_context(
    task: ContextType | List[ContextType] | None,
    context: ContextType | List[ContextType] | None,
) -> ContextType | List[ContextType] | None:
    if task is None:
        return context

    task_list = task if isinstance(task, list) else [task]
    if context is None:
        return task_list

    context_list = context if isinstance(context, list) else [context]
    return [*context_list, *task_list]  # type: ignore[return-value]


def _validate_target_result(
    target: TargetParam[Output] | None, result: Any
) -> Output | None:
    if target is None:
        return None

    if isinstance(target, Target):
        target_value = target.target
    else:
        target_value = target

    builder: OutputBuilder[Any] = OutputBuilder(target=target_value)
    builder.update(result)
    return builder.finalize()


def _evaluate_verifier(
    verifier: RunVerifier | None,
    completion: RunCompletion,
    context: RunVerificationContext[Output],
) -> tuple[bool | None, str | None]:
    if verifier is None:
        return None, None

    try:
        takes_ctx = len(inspect.signature(verifier).parameters) > 1
        if takes_ctx:
            verdict = verifier(completion, context)  # type: ignore[arg-type]
        else:
            verdict = verifier(completion)  # type: ignore[arg-type]
    except Exception as exc:
        return False, str(exc)

    if verdict is None or verdict is True:
        return True, None
    if verdict is False:
        return False, "Verification failed."
    if isinstance(verdict, str):
        return False, verdict
    return False, "Verification failed."


def _evaluate_checks(
    checks: List[RunCheck] | None,
    completion: RunCompletion,
    context: RunVerificationContext[Output],
) -> tuple[bool | None, str | None]:
    if not checks:
        return None, None
    for check in checks:
        try:
            takes_ctx = len(inspect.signature(check).parameters) > 1
            if takes_ctx:
                verdict = check(completion, context)  # type: ignore[arg-type]
            else:
                verdict = check(completion)  # type: ignore[arg-type]
        except Exception as exc:
            return False, str(exc)

        if verdict is None or verdict is True:
            continue
        if verdict is False:
            return False, "Final answer check failed."
        if isinstance(verdict, str):
            return False, verdict
    return True, None


def _normalize_attachments_list(
    attachments: AttachmentType | List[AttachmentType] | None,
) -> List[AttachmentType] | None:
    if attachments is None:
        return None
    if isinstance(attachments, list):
        return cast(List[AttachmentType], attachments)
    return [attachments]


def _apply_completion(
    *,
    state: RunExecutionState[Output],
    summary: str | None,
    status: Literal["success", "failed", "skipped"],
    result: Any | None,
    evidence: Any | None,
    verification: Any | None,
    target: TargetParam[Output] | None,
    source: SourceParam | None,
    deps: Deps | None,
    attachments: AttachmentType | List[AttachmentType] | None,
    verifier: RunVerifier | None,
    final_answer_checks: List[RunCheck] | None,
    error_cls: type[Exception],
) -> None:
    if target is not None and result is None:
        raise error_cls(
            "Task completion requires a result that matches the target schema."
        )

    if result is not None:
        if target is None:
            state.validated_output = result
        else:
            try:
                state.validated_output = _validate_target_result(
                    target, result
                )
            except Exception as exc:
                raise error_cls(
                    f"Task completion result failed validation: {exc}"
                ) from exc

    state.completion_state = RunCompletion(
        status=status,
        summary=summary or "Completed.",
        result=result,
        evidence=evidence,
        verification=verification,
    )

    verification_context = RunVerificationContext(
        target=target,
        source=source,
        deps=deps,  # type: ignore[arg-type]
        attachments=_normalize_attachments_list(attachments),
    )

    state.verified, state.verification_error = _evaluate_verifier(
        verifier, state.completion_state, verification_context
    )

    if state.verified is False:
        raise error_cls(
            state.verification_error or "Verification failed. Try again."
        )

    checks_ok, checks_error = _evaluate_checks(
        final_answer_checks, state.completion_state, verification_context
    )
    if checks_ok is False:
        raise error_cls(
            checks_error or "Final answer checks failed. Try again."
        )


async def _execute_plan_based_run(
    *,
    ctx: Any,
    graph_deps: SemanticGraphDeps[Deps, Output],
    run_state: SimpleNamespace,
    execution_state: RunExecutionState[Output],
    system_prompt: str,
    target: TargetParam[Output] | None,
    source: SourceParam | None,
    deps: Deps | None,
    attachments: AttachmentType | List[AttachmentType] | None,
    verifier: RunVerifier | None,
    max_retries: int,
    planning_interval: int | None,
    max_steps: int | None,
    final_answer_checks: List[RunCheck] | None,
    require_completion: bool,
) -> None:
    observe = getattr(graph_deps, "observe", None)
    tool_inventory = _render_tool_inventory(graph_deps.toolsets)
    if tool_inventory:
        system_prompt += tool_inventory

    plan_prompt = _PLAN_SYSTEM_PROMPT
    if planning_interval is not None:
        plan_prompt += (
            f"- Limit this plan to at most {planning_interval} tool calls.\n"
        )

    plan_rounds = max(1, max_retries)
    last_plan_error: str | None = None
    for round_idx in range(plan_rounds):
        plan_error: str | None = last_plan_error
        plan_code: str | None = None

        for _ in range(max_retries):
            retry_note = (
                f"\n[PLAN ERROR]\n{plan_error}\n[END PLAN ERROR]\n"
                if plan_error
                else ""
            )
            plan_request = SemanticGraphRequestTemplate(
                system_prompt_additions=system_prompt
                + retry_note
                + plan_prompt,
                output_type=str,
                native_output=False,
                include_output_context=False,
                include_user_toolsets=False,
                toolsets=[],
            )
            plan_params = plan_request.render(ctx)
            _observe_model_request(observe)
            plan_result = await graph_deps.agent.run(**plan_params)
            run_state.agent_runs.append(plan_result)
            candidate_code: str = plan_result.output
            candidate_code = candidate_code.strip()
            if candidate_code.startswith("```"):
                candidate_code = candidate_code.strip("`")
                if candidate_code.startswith("python"):
                    candidate_code = candidate_code[len("python") :].strip()

            if "complete_task" in candidate_code:
                candidate_code = candidate_code.replace(
                    "complete_task", "complete"
                )

            forbidden = ("await ", "multi_tool_use", "functions.")
            bad_token = next(
                (token for token in forbidden if token in candidate_code),
                None,
            )
            if bad_token:
                plan_error = (
                    f"Plan contains forbidden token: {bad_token!r}. "
                    "Plan must be pure Python code with direct tool calls."
                )
                continue

            plan_code = candidate_code
            plan_error = None
            break

        if plan_code is None:
            raise ValueError(
                f"Failed to generate a valid plan after {max_retries} attempts."
            )

        prompt_blocks: List[str] = []
        code_blocks: List[str] = [plan_code]

        tool_funcs: Dict[str, Callable[..., Any]] = {}

        def _wrap_tool(name: str, tool: Any) -> Callable[..., Any]:
            def _wrapped(*args, **kwargs):
                execution_state.tool_call_log[name] = (
                    execution_state.tool_call_log.get(name, 0) + 1
                )
                execution_state.total_tool_calls += 1
                if (
                    max_steps is not None
                    and execution_state.total_tool_calls > max_steps
                ):
                    raise ValueError(f"Max steps exceeded ({max_steps}).")

                if not isinstance(graph_deps.agent.model, PydanticAIModel):
                    model = graph_deps.agent._get_model(
                        model=graph_deps.agent.model,
                    )
                else:
                    model = graph_deps.agent.model

                ctx = PydanticAIRunContext(
                    deps=graph_deps.deps,
                    model=model,
                    usage=PydanticAIUsage(),
                )
                if getattr(tool, "takes_ctx", False):
                    return tool.function(ctx, *args, **kwargs)
                return tool.function(*args, **kwargs)

            return _wrapped

        for toolset_item in graph_deps.toolsets:
            if isinstance(toolset_item, FunctionToolset):
                for name, tool in toolset_item.tools.items():
                    tool_funcs[name] = _wrap_tool(name, tool)

        def tool_called(name: str) -> bool:
            return execution_state.tool_call_log.get(name, 0) > 0

        def tool_calls(name: str | None = None) -> int | Dict[str, int]:
            if name is None:
                return dict(execution_state.tool_call_log)
            return execution_state.tool_call_log.get(name, 0)

        def prompt(text: str) -> None:
            prompt_blocks.append(text)

        def complete(
            summary: str | None = None,
            *,
            status: Literal["success", "failed", "skipped"] = "success",
            result: Any | None = None,
            evidence: Any | None = None,
            verification: Any | None = None,
        ) -> None:
            _apply_completion(
                state=execution_state,
                summary=summary,
                status=status,
                result=result,
                evidence=evidence,
                verification=verification,
                target=target,
                source=source,
                deps=deps,
                attachments=attachments,
                verifier=verifier,
                final_answer_checks=final_answer_checks,
                error_cls=ValueError,
            )

        external_functions = {
            **tool_funcs,
            "tool_called": tool_called,
            "tool_calls": tool_calls,
            "prompt": prompt,
            "complete": complete,
            "complete_task": complete,
        }

        plan_exec_error: str | None = None
        round_start_calls = execution_state.total_tool_calls
        for _ in range(max_retries):
            try:
                for code in code_blocks:
                    monty = pydantic_monty.Monty(
                        code,
                        inputs=[],
                        external_functions=list(external_functions.keys()),
                        script_name="run_plan.py",
                        type_check=False,
                    )
                    await pydantic_monty.run_monty_async(
                        monty,
                        external_functions=external_functions,
                    )
                if planning_interval is not None:
                    round_calls = (
                        execution_state.total_tool_calls - round_start_calls
                    )
                    if round_calls > planning_interval:
                        raise ValueError(
                            f"Plan exceeded tool-call limit ({planning_interval})."
                        )
                plan_exec_error = None
                break
            except Exception as exc:
                plan_exec_error = str(exc)
                continue

        if plan_exec_error:
            last_plan_error = f"Plan execution failed: {plan_exec_error}"
            if round_idx < plan_rounds - 1:
                continue
            raise ValueError(
                f"Plan execution failed after {plan_rounds} plan rounds: {plan_exec_error}"
            )

        if prompt_blocks:
            prompt_request = SemanticGraphRequestTemplate(
                system_prompt_additions=system_prompt,
                user_prompt_additions="\n\n".join(prompt_blocks),
                output_type=None,
                native_output=False,
                include_output_context=False,
                include_user_toolsets=False,
                toolsets=[],
            )
            prompt_params = prompt_request.render(ctx)
            _observe_model_request(observe)
            prompt_result = await graph_deps.agent.run(**prompt_params)
            run_state.agent_runs.append(prompt_result)

        if execution_state.completion_state is not None:
            break
        last_plan_error = "Plan did not complete the task."

    if require_completion and execution_state.completion_state is None:
        raise ValueError(
            f"Task did not complete after {plan_rounds} plan rounds."
        )


async def _execute_standard_run(
    *,
    ctx: Any,
    graph_deps: SemanticGraphDeps[Deps, Output],
    run_state: SimpleNamespace,
    execution_state: RunExecutionState[Output],
    system_prompt: str,
    toolset: FunctionToolset,
    require_completion: bool,
    max_retries: int,
    max_steps: int | None,
) -> None:
    observe = getattr(graph_deps, "observe", None)
    request = SemanticGraphRequestTemplate(
        system_prompt_additions=system_prompt,
        output_type=None,
        native_output=False,
        include_output_context=False,
        toolsets=[toolset],
    )
    last_error: str | None = None
    max_attempts = max_steps if max_steps is not None else max_retries
    for _ in range(max_attempts):
        if last_error:
            request.user_prompt_additions = (
                f"Previous attempt failed: {last_error}"
            )
        params = request.render(ctx)
        _observe_model_request(observe)
        result = await graph_deps.agent.run(**params)

        run_state.agent_runs.append(result)

        if require_completion and execution_state.completion_state is None:
            last_error = (
                "Task did not complete. You must call `complete_task`."
            )
            continue
        last_error = None
        break

    if last_error:
        raise ValueError(last_error)


async def _execute_stream_run(
    *,
    ctx: Any,
    graph_deps: SemanticGraphDeps[Deps, Output],
    run_state: SimpleNamespace,
    execution_state: RunExecutionState[Output],
    system_prompt: str,
    toolset: FunctionToolset,
    require_completion: bool,
    target: TargetParam[Output],
    task_goal: str | None,
    source: SourceParam | None,
    attachments: AttachmentType | List[AttachmentType] | None,
) -> RunStream[Output]:
    observe = getattr(graph_deps, "observe", None)
    request = SemanticGraphRequestTemplate(
        system_prompt_additions=system_prompt,
        output_type=graph_deps.target_strategy.target
        if graph_deps.target_strategy
        else None,
        native_output=False,
        include_output_context=False,
        toolsets=[toolset],
    )
    params = request.render(ctx)
    handler = (
        getattr(observe, "event_stream_handler", None) if observe else None
    )
    _observe_model_request(observe)
    stream_ctx = await graph_deps.agent.run_stream(
        **params, event_stream_handler=handler
    )
    stream = await stream_ctx.__aenter__()
    builder = OutputBuilder(
        target=graph_deps.target_strategy.target
        if graph_deps.target_strategy
        else target
    )
    stream_wrapper = RunStream(
        _builder=builder,
        _streams=[stream],
        _field_mappings=[
            StreamStreamFieldMapping(
                stream_index=0, fields=None, update_output=True
            )
        ],
        _stream_contexts=[stream_ctx],
        completion=execution_state.completion_state,
        _exclude_none=False,
        verified=execution_state.verified,
        verification_error=execution_state.verification_error,
        task=RunTask(goal=task_goal, target=target, source=source),
        _require_completion=require_completion,
        _execution_state=execution_state,
    )
    return stream_wrapper


async def _run_async(
    task: ContextType | List[ContextType] | None,
    target: TargetParam[Output] | None,
    context: ContextType | List[ContextType] | None,
    *,
    source: SourceParam | None,
    instructions: PydanticAIInstructions | None,
    model: ModelParam,
    model_settings: PydanticAIModelSettings | None,
    attachments: AttachmentType | List[AttachmentType] | None,
    tools: ToolType | List[ToolType] | None,
    deps: Deps | None,
    usage_limits: PydanticAIUsageLimits | None,
    verifier: RunVerifier | None,
    require_completion: bool,
    plan: bool,
    max_retries: int,
    planning_interval: int | None,
    max_steps: int | None,
    final_answer_checks: List[RunCheck] | None,
    observe: bool | "Observer" | None,
    stream: bool = False,
) -> RunResult[Output] | RunStream[Output]:
    if task is None:
        make_target = cast(
            TargetParam[Output], target if target is not None else str
        )
        return await amake(  # type: ignore[no-matching-overload]
            target=make_target,
            context=context,
            instructions=instructions,
            model=model,
            model_settings=model_settings,
            attachments=attachments,
            tools=tools,
            deps=deps,
            usage_limits=usage_limits,
            observe=_normalize_observer(observe),
            stream=stream,
        )

    merged_context = _normalize_context(task=task, context=context)

    _target = target
    _instructions = instructions
    if isinstance(target, Target):
        _target = target.target
        if target.instructions and not _instructions:
            _instructions = target.instructions

    graph_deps = SemanticGraphDeps.prepare(
        model=model,
        model_settings=model_settings,
        context=merged_context,
        instructions=_instructions,  # type: ignore[arg-type]
        tools=tools,
        deps=deps,
        attachments=attachments,
        usage_limits=usage_limits,
        target=_target,
        source=source,
        inject_internal_deps=True,
        observe=_normalize_observer(observe),
    )

    execution_state = RunExecutionState[Output]()

    toolset = FunctionToolset()

    @toolset.tool
    def complete_task(
        summary: str,
        *,
        status: Literal["success", "failed", "skipped"] = "success",
        result: Any | None = None,
        evidence: Any | None = None,
        verification: Any | None = None,
    ) -> str:
        _apply_completion(
            state=execution_state,
            summary=summary,
            status=status,
            result=result,
            evidence=evidence,
            verification=verification,
            target=target,
            source=source,
            deps=deps,
            attachments=attachments,
            verifier=verifier,
            final_answer_checks=final_answer_checks,
            error_cls=ModelRetry,
        )

        return "ok"

    system_prompt = _RUN_SYSTEM_PROMPT
    target_schema = _build_target_schema_text(target)
    if target_schema:
        system_prompt += (
            "\n[EXPECTED RESULT SCHEMA]\n"
            f"{target_schema}\n"
            "[END EXPECTED RESULT SCHEMA]\n"
            "- When calling `complete_task`, ensure `result` matches the schema.\n"
        )
    else:
        system_prompt += (
            "\n- No result schema is required. Use `result=null` unless the task explicitly "
            "requests a specific output.\n"
        )

    run_state = SimpleNamespace(agent_runs=[], output=None)
    ctx = SimpleNamespace(state=run_state, deps=graph_deps)

    if stream and target is None:
        raise ValueError("`stream=True` requires `target` to be provided.")

    if stream and plan:
        raise ValueError("`stream=True` is not supported with `plan=True`.")

    observe = getattr(graph_deps, "observe", None)
    if observe:
        observe.on_operation_start("run")

    try:
        if plan:
            await _execute_plan_based_run(
                ctx=ctx,
                graph_deps=graph_deps,
                run_state=run_state,
                execution_state=execution_state,
                system_prompt=system_prompt,
                target=target,
                source=source,
                deps=deps,
                attachments=attachments,
                verifier=verifier,
                max_retries=max_retries,
                planning_interval=planning_interval,
                max_steps=max_steps,
                final_answer_checks=final_answer_checks,
                require_completion=require_completion,
            )
        else:
            if stream and target is not None:
                task_goal: str | None = None
                if isinstance(task, str):
                    task_goal = task
                return await _execute_stream_run(
                    ctx=ctx,
                    graph_deps=graph_deps,
                    run_state=run_state,
                    execution_state=execution_state,
                    system_prompt=system_prompt,
                    toolset=toolset,
                    require_completion=require_completion,
                    target=target,
                    task_goal=task_goal,
                    source=source,
                    attachments=attachments,
                )
            await _execute_standard_run(
                ctx=ctx,
                graph_deps=graph_deps,
                run_state=run_state,
                execution_state=execution_state,
                system_prompt=system_prompt,
                toolset=toolset,
                require_completion=require_completion,
                max_retries=max_retries,
                max_steps=max_steps,
            )
    finally:
        if observe:
            observe.on_operation_complete("run")

    if require_completion and execution_state.completion_state is None:
        raise ValueError(
            "Task did not complete. The model must call `complete`."
        )

    completion_state = execution_state.completion_state or RunCompletion(
        status="skipped",
        summary="No completion tool call was made.",
    )

    task_goal: str | None = None
    if isinstance(task, str):
        task_goal = task

    ctx_refs = getattr(graph_deps, "_context_refs", None) or []
    if not isinstance(ctx_refs, list):
        ctx_refs = [ctx_refs]
    context_additions = getattr(graph_deps, "_context_additions", []) or []
    for ctx_ref in ctx_refs:
        if ctx_ref is None or not getattr(ctx_ref, "update", True):
            continue
        if context_additions:
            ctx_ref.extend_messages(context_additions)
        ctx_ref.add_assistant_message(
            semantic_for_operation(
                "run",
                summary=completion_state.summary,
                output=execution_state.validated_output,
            )
        )

    return RunResult(
        output=execution_state.validated_output,
        raw=run_state.agent_runs,
        completion=completion_state,
        verified=execution_state.verified,
        verification_error=execution_state.verification_error,
        task=RunTask(goal=task_goal, target=target, source=source),
    )


@overload
async def arun(
    task: ContextType | List[ContextType] | None = ...,
    target: TargetParam[Output] | None = ...,
    context: ContextType | List[ContextType] | None = ...,
    *,
    source: SourceParam | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    attachments: AttachmentType | List[AttachmentType] | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    verifier: RunVerifier | None = ...,
    require_completion: bool = ...,
    plan: bool = ...,
    max_retries: int = ...,
    planning_interval: int | None = ...,
    max_steps: int | None = ...,
    final_answer_checks: List[RunCheck] | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[True],
) -> RunStream[Output]: ...


@overload
async def arun(
    task: ContextType | List[ContextType] | None = ...,
    target: TargetParam[Output] | None = ...,
    context: ContextType | List[ContextType] | None = ...,
    *,
    source: SourceParam | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    attachments: AttachmentType | List[AttachmentType] | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    verifier: RunVerifier | None = ...,
    require_completion: bool = ...,
    plan: bool = ...,
    max_retries: int = ...,
    planning_interval: int | None = ...,
    max_steps: int | None = ...,
    final_answer_checks: List[RunCheck] | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[False] = False,
) -> RunResult[Output]: ...


async def arun(
    task: ContextType | List[ContextType] | None = None,
    target: TargetParam[Output] | None = None,
    context: ContextType | List[ContextType] | None = None,
    *,
    source: SourceParam | None = None,
    instructions: PydanticAIInstructions | None = None,
    model: ModelParam = "openai:gpt-4o-mini",
    model_settings: PydanticAIModelSettings | None = None,
    attachments: AttachmentType | List[AttachmentType] | None = None,
    tools: ToolType | List[ToolType] | None = None,
    deps: Deps | None = None,
    usage_limits: PydanticAIUsageLimits | None = None,
    verifier: RunVerifier | None = None,
    require_completion: bool = True,
    plan: bool = False,
    max_retries: int = 2,
    planning_interval: int | None = None,
    max_steps: int | None = None,
    final_answer_checks: List[RunCheck] | None = None,
    observe: bool | Observer | None = None,
    stream: bool = False,
) -> RunResult[Output] | RunStream[Output]:
    """Asynchronously run a task with tool-verified completion.

    `run()` is goal-centric: completion is signaled by a tool call, not by a
    typed output. If `target` is provided, the completion tool must include
    a result that validates against the target schema.
    """
    return await _run_async(
        task=task,
        target=target,
        context=context,
        source=source,
        instructions=instructions,
        model=model,
        model_settings=model_settings,
        attachments=attachments,
        tools=tools,
        deps=deps,
        usage_limits=usage_limits,
        verifier=verifier,
        require_completion=require_completion,
        plan=plan,
        max_retries=max_retries,
        planning_interval=planning_interval,
        max_steps=max_steps,
        final_answer_checks=final_answer_checks,
        observe=observe,
        stream=stream,
    )


@overload
def run(
    task: ContextType | List[ContextType] | None = ...,
    target: TargetParam[Output] | None = ...,
    context: ContextType | List[ContextType] | None = ...,
    *,
    source: SourceParam | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    attachments: AttachmentType | List[AttachmentType] | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    verifier: RunVerifier | None = ...,
    require_completion: bool = ...,
    plan: bool = ...,
    max_retries: int = ...,
    planning_interval: int | None = ...,
    max_steps: int | None = ...,
    final_answer_checks: List[RunCheck] | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[True],
) -> RunStream[Output]: ...


@overload
def run(
    task: ContextType | List[ContextType] | None = ...,
    target: TargetParam[Output] | None = ...,
    context: ContextType | List[ContextType] | None = ...,
    *,
    source: SourceParam | None = ...,
    instructions: PydanticAIInstructions | None = ...,
    model: ModelParam = ...,
    model_settings: PydanticAIModelSettings | None = ...,
    attachments: AttachmentType | List[AttachmentType] | None = ...,
    tools: ToolType | List[ToolType] | None = ...,
    deps: Deps | None = ...,
    usage_limits: PydanticAIUsageLimits | None = ...,
    verifier: RunVerifier | None = ...,
    require_completion: bool = ...,
    plan: bool = ...,
    max_retries: int = ...,
    planning_interval: int | None = ...,
    max_steps: int | None = ...,
    final_answer_checks: List[RunCheck] | None = ...,
    observe: bool | Observer | None = ...,
    stream: Literal[False] = False,
) -> RunResult[Output]: ...


def run(
    task: ContextType | List[ContextType] | None = None,
    target: TargetParam[Output] | None = None,
    context: ContextType | List[ContextType] | None = None,
    *,
    source: SourceParam | None = None,
    instructions: PydanticAIInstructions | None = None,
    model: ModelParam = "openai:gpt-4o-mini",
    model_settings: PydanticAIModelSettings | None = None,
    attachments: AttachmentType | List[AttachmentType] | None = None,
    tools: ToolType | List[ToolType] | None = None,
    deps: Deps | None = None,
    usage_limits: PydanticAIUsageLimits | None = None,
    verifier: RunVerifier | None = None,
    require_completion: bool = True,
    plan: bool = False,
    max_retries: int = 2,
    planning_interval: int | None = None,
    max_steps: int | None = None,
    final_answer_checks: List[RunCheck] | None = None,
    observe: bool | Observer | None = None,
    stream: bool = False,
) -> RunResult[Output] | RunStream[Output]:
    """Run a task with tool-verified completion."""
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = None

    if loop and loop.is_running():
        raise RuntimeError(
            "Cannot call `run()` inside an async context. Use `await arun()`."
        )

    return asyncio.run(
        _run_async(
            task=task,
            target=target,
            context=context,
            source=source,
            instructions=instructions,
            model=model,
            model_settings=model_settings,
            attachments=attachments,
            tools=tools,
            deps=deps,
            usage_limits=usage_limits,
            verifier=verifier,
            require_completion=require_completion,
            plan=plan,
            max_retries=max_retries,
            planning_interval=planning_interval,
            max_steps=max_steps,
            final_answer_checks=final_answer_checks,
            observe=observe,
            stream=stream,
        )
    )



"""
scripts.run_readme_examples

Tests/validates all examples provided within the
readme.md file.

This also ensures that all examples are containerized in
terms of dependencies.
"""


def run_generate_content_example() -> None:
    import zyx

    result = zyx.make(
        target=int,
        context="What is 45+45?",
        model="openai:gpt-4o-mini",
    )

    print(result.output)


def run_parse_structured_data_example() -> None:
    import zyx
    from pydantic import BaseModel

    class Information(BaseModel):
        library_name: str
        library_description: str

    result = zyx.parse(
        source=zyx.paste("https://zyx.hammad.app"),
        target=Information,
        model="openai:gpt-4o-mini",
    )

    print(result.output.library_name)
    print(result.output.library_description)


def run_query_with_tools_example() -> None:
    import zyx
    from pydantic import BaseModel

    class Information(BaseModel):
        library_name: str
        library_description: str

    def log_website_url(url: str) -> None:
        print(f"Website URL: {url}")

    result = zyx.parse(
        source=zyx.paste("https://zyx.hammad.app"),
        target=Information,
        context=[
            {"role": "system", "content": "You are a web scraper."},
            # NOTE:
            # should test with a mock file at some point
            # zyx.paste("scraping_instructions.txt"),
            "[s]log the website URL before you parse.[/s]",
        ],
        model="openai:gpt-4o-mini",
        tools=[log_website_url],
    )

    print(result.output.library_name)
    print(result.output.library_description)


def run_edit_values_example() -> None:
    import zyx

    data = {"name": "John", "age": 30}

    result = zyx.edit(
        target=data,
        context="Update the age to 31",
        model="openai:gpt-4o-mini",
        merge=False
    )

    print(result.output)


def run_query_grounded_sources_example() -> None:
    import zyx

    result = zyx.query(
        source="Python is a high-level programming language...",
        target=str,
        context="What is Python?",
        model="openai:gpt-4o-mini",
    )

    print(result.output)


def run_select_from_options_example() -> None:
    import zyx
    from typing import Literal

    Color = Literal["red", "green", "blue"]

    result = zyx.select(
        target=Color,
        context="What color is the sky?",
        model="openai:gpt-4o-mini",
    )

    print(result.output)


def run_async_support_example() -> None:
    import asyncio

    async def test() -> None:
        import zyx

        result = await zyx.amake(
            target=str,
            context="Write a haiku about Python",
            model="openai:gpt-4o-mini",
        )

        print(result.output)

    asyncio.run(test())


def run_streaming_support_example() -> None:
    import zyx

    stream = zyx.make(
        target=str,
        context="Write a short story",
        model="openai:gpt-4o-mini",
        stream=True,
    )

    for chunk in stream.text(delta=True):
        print(chunk)


if __name__ == "__main__":
    # run_generate_content_example()
    # run_parse_structured_data_example()
    # run_query_with_tools_example()
    # run_edit_values_example()
    # run_query_grounded_sources_example()
    # run_select_from_options_example()
    # run_async_support_example()
    run_streaming_support_example()


"""zyx._utils._strategies._params

Framework specific strategies for the `source` and
`target` parameters of semantic operations.
"""

from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Literal

from pydantic_ai.messages import UserPromptPart

from ._abstract import AbstractStrategy
from ..._aliases import PydanticAIModelRequest
from ..._processing._multimodal import (
    MultimodalContentOrigin,
    MultimodalContentMediaType,
    render_multimodal_source_as_text,
    render_multimodal_source_as_user_content,
    render_multimodal_source_as_description,
)
from ..._processing._toon import object_as_text
from ...targets import Target
from ...attachments import Attachment


@dataclass
class SourcePayload:
    kind: Literal["raw", "multimodal"]
    content: str
    origin: MultimodalContentOrigin | None = None
    media_type: MultimodalContentMediaType | None = None
    source_repr: str | None = None
    message: PydanticAIModelRequest | None = None

    @property
    def is_text_based(self) -> bool:
        if self.media_type is None:
            return True
        return self.media_type in (
            MultimodalContentMediaType.TEXT,
            MultimodalContentMediaType.HTML,
            MultimodalContentMediaType.DOCUMENT,
            MultimodalContentMediaType.UNKNOWN,
        )


@dataclass
class SourceStrategy(AbstractStrategy):
    source: Any

    @property
    def kind(self) -> Literal["source"]:
        return "source"

    def get_payload(self) -> SourcePayload:
        if isinstance(self.source, Attachment):
            description = self.source.description
            content = self.source.text
            if content:
                if description:
                    content = f"{description}\n\n{content}"
                return SourcePayload(kind="raw", content=content)
            message = self.source.message
            return SourcePayload(
                kind="multimodal",
                content=description,
                message=message,
            )

        if isinstance(self.source, (str, Path)):
            path = Path(self.source)
            if path.exists() and path.is_file():
                origin = MultimodalContentOrigin.classify(path)
                media_type = MultimodalContentMediaType.classify(path, origin)
                is_text_based = media_type in (
                    MultimodalContentMediaType.TEXT,
                    MultimodalContentMediaType.HTML,
                    MultimodalContentMediaType.DOCUMENT,
                    MultimodalContentMediaType.UNKNOWN,
                )
                text = render_multimodal_source_as_text(
                    path, origin, media_type
                )
                description = render_multimodal_source_as_description(
                    path, origin, media_type
                )
                content = text if is_text_based else description
                message = None
                if not is_text_based:
                    user_content = render_multimodal_source_as_user_content(
                        path, origin, media_type
                    )
                    message = PydanticAIModelRequest(
                        parts=[UserPromptPart(content=[user_content])]
                    )
                return SourcePayload(
                    kind="multimodal",
                    content=content,
                    origin=origin,
                    media_type=media_type,
                    source_repr=str(path),
                    message=message,
                )

        return SourcePayload(
            kind="raw",
            content=object_as_text(self.source),
        )


@dataclass
class TargetStrategy(AbstractStrategy):
    target: Any
    name: str | None = None
    description: str | None = None
    instructions: Any | None = None
    has_hooks: bool = False
    is_target_wrapper: bool = False

    @property
    def kind(self) -> Literal["target"]:
        return "target"

    @classmethod
    def from_target(cls, target: Any) -> "TargetStrategy":
        if isinstance(target, Target):
            has_hooks = bool(
                target._field_hooks
                or target._prebuilt_hooks.get("complete", [])
            )
            return cls(
                target=target.target,
                name=target.name,
                description=target.description,
                instructions=target.instructions,
                has_hooks=has_hooks,
                is_target_wrapper=True,
            )
        return cls(target=target)



"""zyx._utils._strategies._editable"""

from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
import json
from typing import Any, List, Literal, Type, Generic, TypeVar

from pydantic import BaseModel, Field, create_model

from ._abstract import AbstractStrategy
from ..._processing._outputs import (
    partial_output_model,
    sparse_output_model,
    selection_output_model,
    split_output_model_by_fields,
    split_output_model,
)
from .._outputs import OutputBuilder


Output = TypeVar("Output")

SHORT_TEXT_REPLACEMENT_THRESHOLD = 75


@dataclass
class AbstractEditableStrategy(AbstractStrategy, ABC, Generic[Output]):
    """
    Abstract base class for strategies referring to objects that can be
    edited through natural language instructions / or the `edit`
    semantic operation.
    """

    class Confirmation(BaseModel):
        needs_edits: bool = Field(
            ...,
            description="Determination if any edits need to be made to the source data provided to you.",
        )

    builder: OutputBuilder[Output]

    @abstractmethod
    def get_replacement_schema(self) -> Any:
        """
        Returns the schema/model describing a full replacement for the editable object.
        """
        raise NotImplementedError

    @abstractmethod
    def get_selective_schema(self) -> Any:
        """
        Returns the schema/model describing a set of selective (field- or anchor-specific) edits.
        """
        raise NotImplementedError

    @abstractmethod
    def get_plan_schema(self) -> Any:
        """
        Returns the schema/model describing an edit 'plan' (e.g., anchors/selections to be edited).
        """
        raise NotImplementedError

    @abstractmethod
    def get_plan_edits_schema(self, plan: Any) -> list[Any]:
        """
        Given a plan, returns a list of schemas/models describing the possible edits for each selection
        in the plan.
        """
        raise NotImplementedError

    @abstractmethod
    def get_plan_edit_schema(self, plan: Any) -> Any:
        """
        Given a plan, returns a schema/model with enumerated fields for each edit to be performed
        according to the selections in the plan.
        """
        raise NotImplementedError

    @abstractmethod
    def apply_replacement(self, replacement: Any) -> Any:
        """
        Applies a full-object replacement, returning the updated value.
        """
        raise NotImplementedError

    @abstractmethod
    def apply_selective(self, edits: Any) -> Any:
        """
        Applies selective (field- or anchor-specific) edits provided by the model,
        returning the updated value.
        """
        raise NotImplementedError

    @abstractmethod
    def apply_plan_edits(self, plan: Any, edits: Any) -> Any:
        """
        Applies a set of planned edits to the object based on the provided plan (e.g., anchor selections)
        and corresponding replacement edits.
        """
        raise NotImplementedError


@dataclass
class BasicEditStrategy(AbstractEditableStrategy[Output]):
    """
    Fallback editable strategy for non-text, non-mapping outputs.
    """

    @property
    def kind(self) -> Literal["basic"]:
        return "basic"

    def get_replacement_schema(self) -> Type[BaseModel]:
        normalized = self.builder.partial_type
        model = create_model(
            "Replacement",
            value=(
                normalized,
                Field(
                    ...,
                    description=(
                        "A complete replacement value for the data provided to you. "
                        "Return the fully updated value here."
                    ),
                ),
            ),
        )
        return model

    def get_selective_schema(self) -> Type[BaseModel]:
        return self.get_replacement_schema()

    def get_plan_schema(self) -> Type[BaseModel]:
        Plan = create_model(
            "EditPlan",
            selection=(
                Literal["null", "replace"],
                Field(
                    ...,
                    description=(
                        "Plan for editing this value. "
                        "'null' = no changes are needed. "
                        "'replace' = you will provide a complete replacement value."
                    ),
                ),
            ),
        )
        return Plan

    def get_plan_edits_schema(self, plan: BaseModel) -> List[Type[BaseModel]]:
        if getattr(plan, "selection") == "replace":
            return [self.get_replacement_schema()]
        return []

    def get_plan_edit_schema(self, plan: BaseModel) -> Type[BaseModel]:
        return self.get_replacement_schema()

    def apply_replacement(self, replacement: Any) -> Output:
        value = getattr(replacement, "value", replacement)
        if (
            self.builder.normalized is str
            and isinstance(value, str)
            and value.startswith('"')
            and value.endswith('"')
        ):
            try:
                decoded = json.loads(value)
            except json.JSONDecodeError:
                decoded = value
            else:
                if isinstance(decoded, str):
                    value = decoded
        return self.builder.update(value)

    def apply_selective(self, edits: Any) -> Output:
        return self.apply_replacement(edits)

    def apply_plan_edits(self, plan: BaseModel, edits: Any) -> Output:
        if getattr(plan, "selection", None) == "null":
            if self.builder.partial is None and self.builder.is_value:
                return self.builder.update(self.builder.target)
            return self.builder.partial  # type: ignore[return-value]
        return self.apply_replacement(edits)


@dataclass
class TextEditStrategy(AbstractEditableStrategy[str]):
    """Strategy for edits made to strings."""

    def __post_init__(self) -> None:
        if self.builder.normalized is not str:
            raise ValueError(
                "TextEditStrategy can only be used with string targets."
            )

    class Replacement(BaseModel):
        """A complete replacement for a piece of text/a string. All content
        within the text or string will be replaced with this replacement."""

        text: str = Field(
            ..., description="The text/content to replace the section with."
        )

    class Edit(BaseModel):
        """A single edit operation to be applied onto a string.

        This represents a single edit, and must include a start and end
        anchor which represents the starting and ending snippets of the
        section that is being edited."""

        start_anchor: str = Field(
            ...,
            description="The starting text/content of the section that is being edited.",
        )

        end_anchor: str = Field(
            ...,
            description="The ending text/content of the section that is being edited.",
        )

        replacement: str = Field(
            ..., description="The text/content to replace the section with."
        )

    class Edits(BaseModel):
        """A plan for a series of edit operations to be applied onto text
        content."""

        edits: List[TextEditStrategy.Edit] = Field(
            ...,
            description="The list of edit operations to be applied onto the text content.",
        )

    class Selection(BaseModel):
        """A selection of anchors to be edited within a string/text content. This
        is used to mark the starting and ending snippets of a specific section of
        content that is being edited."""

        start_anchor: str = Field(
            ...,
            description="The starting text/content of the section that is being edited.",
        )

        end_anchor: str = Field(
            ...,
            description="The ending text/content of the section that is being edited.",
        )

    class EditPlan(BaseModel):
        """A plan for a series of edit operations to be applied onto text content. Once
        you have selected the appropriate anchors for the context that will be replaced,
        you will be able to provide a replacement for the selected section(s).

        If not content needs to be edited, or replaced only return the 'null' option."""

        selections: List[TextEditStrategy.Selection] | Literal["null"] = Field(
            ...,
            description="The list of selections of anchors to be edited within the text content.",
        )

    @property
    def kind(self) -> Literal["text"]:
        return "text"

    def get_replacement_schema(self) -> Type[BaseModel]:
        return self.Replacement

    def get_selective_schema(self) -> Type[BaseModel]:
        return self.Edits

    def get_plan_schema(self) -> Type[BaseModel]:
        return self.EditPlan

    def get_plan_edit_schema(self, plan: BaseModel | Any) -> Type[BaseModel]:
        fields = {}
        for idx, selection in enumerate(plan.selections, 1):  # type: ignore[union-attr]
            field_name = f"edit_{idx}"
            field_doc = (
                f"Replacement for the selection spanning:\n"
                f"Start anchor (index {idx}): '{selection.start_anchor}'\n"
                f"End anchor (index {idx}): '{selection.end_anchor}'"
            )
            fields[field_name] = (
                self.get_replacement_schema(),
                Field(..., description=field_doc),
            )

        model_name = "Edits"
        docstring = (
            "Enumerated fields ('edit_1', 'edit_2', ...) corresponding to each selection "
            "within the plan. Each field contains a replacement for the section of text "
            "identified by its start/end anchors, as described in the field."
        )
        new_model = create_model(model_name, **fields)
        new_model.__doc__ = docstring
        return new_model

    def get_plan_edits_schema(
        self, plan: BaseModel | Any
    ) -> List[Type[BaseModel]]:
        schemas = []
        for _edit in plan.selections:  # type: ignore
            schemas.append(self.get_replacement_schema())

        return schemas

    def _base_text(self) -> str:
        if self.builder.partial is not None:
            return self.builder.partial
        if self.builder.is_value and isinstance(self.builder.target, str):
            return self.builder.target
        raise ValueError("Text edits require a string target value.")

    def _apply_anchor_edit(
        self,
        text: str,
        *,
        start_anchor: str,
        end_anchor: str,
        replacement: str,
    ) -> str:
        start_idx = text.find(start_anchor)
        if start_idx == -1:
            return text
        end_idx = text.find(end_anchor, start_idx + len(start_anchor))
        if end_idx == -1:
            return text
        end_idx += len(end_anchor)
        return text[:start_idx] + replacement + text[end_idx:]

    def apply_replacement(self, replacement: Any) -> str:
        text = getattr(replacement, "text", replacement)
        return self.builder.update(text)

    def apply_selective(self, edits: Any) -> str:
        base = self._base_text()
        if isinstance(edits, BaseModel) and hasattr(edits, "edits"):
            edit_list = edits.edits
        else:
            edit_list = edits
        updated = base
        for edit in edit_list:
            updated = self._apply_anchor_edit(
                updated,
                start_anchor=edit.start_anchor,
                end_anchor=edit.end_anchor,
                replacement=edit.replacement,
            )
        return self.builder.update(updated)

    def apply_plan_edits(self, plan: BaseModel, edits: Any) -> str:
        selections = getattr(plan, "selections", None)
        if selections in (None, "null"):
            if self.builder.partial is None and self.builder.is_value:
                return self.builder.update(self.builder.target)
            return self.builder.partial  # type: ignore[return-value]

        base = self._base_text()
        updated = base
        if isinstance(edits, BaseModel):
            edits_map = {
                name: getattr(edits, name)
                for name in getattr(edits, "model_fields", {}).keys()
            }
            for idx, selection in enumerate(selections, 1):
                repl = edits_map.get(f"edit_{idx}")
                if repl is None:
                    continue
                replacement_text = getattr(repl, "text", repl)
                updated = self._apply_anchor_edit(
                    updated,
                    start_anchor=selection.start_anchor,
                    end_anchor=selection.end_anchor,
                    replacement=replacement_text,
                )
        else:
            for selection, repl in zip(selections, edits):
                replacement_text = getattr(repl, "text", repl)
                updated = self._apply_anchor_edit(
                    updated,
                    start_anchor=selection.start_anchor,
                    end_anchor=selection.end_anchor,
                    replacement=replacement_text,
                )

        return self.builder.update(updated)

    def apply_plan_edit_at_index(
        self, plan: "TextEditStrategy.EditPlan", index: int, edit: Any
    ) -> str:
        selections = getattr(plan, "selections", None)
        if selections in (None, "null"):
            if self.builder.partial is None and self.builder.is_value:
                return self.builder.update(self.builder.target)
            return self.builder.partial  # type: ignore[return-value]

        if not isinstance(selections, list) or index >= len(selections):
            return self.builder.partial  # type: ignore[return-value]

        selection = selections[index]
        replacement_text = getattr(edit, "text", edit)
        base = self._base_text()
        updated = self._apply_anchor_edit(
            base,
            start_anchor=selection.start_anchor,
            end_anchor=selection.end_anchor,
            replacement=replacement_text,
        )
        return self.builder.update(updated)


class MappingEditStrategy(AbstractEditableStrategy[Output]):
    """Strategy for edits made to dictionaries/pydantic models/other mapping-like
    structures and types."""

    @property
    def kind(self) -> Literal["mapping"]:
        return "mapping"

    def get_replacement_schema(self) -> Type[BaseModel]:
        model = partial_output_model(self.builder.normalized)  # type: ignore[arg-type]
        doc = model.__doc__

        if doc is None:
            doc = ""
        doc += (
            "A complete replacement for mapping-like data that has been provided to you."
            "Only fields that should be replaced should be present."
        )
        model.__doc__ = doc
        model.__name__ = "Replacement"

        return model

    def get_selective_schema(self) -> Type[BaseModel]:
        model = sparse_output_model(self.builder.normalized)
        doc = model.__doc__

        if doc is None:
            doc = ""
        doc += (
            "A selective list of changes to be applied to the mapping-like data that has been provided to you."
            "Only fields that should be changed should be present."
        )
        model.__doc__ = doc
        model.__name__ = "Edits"

        return model

    def get_plan_schema(self) -> Type[BaseModel]:
        selections = self.builder.field_names + ["null", "all"]

        model = selection_output_model(
            selections, multi_label=True, name="EditPlan"
        )
        model.__doc__ = (
            "A plan for a series of edit operations to be applied onto the mapping-like data that has been provided to you."
            "Provide only the names of the fields that should be edited or replaced. If a field can be left as None, or as is, do not"
            "include it in the plan."
            "If no fields should be edited, selected, or replaced, select the 'null' option, and vice-versa with the `all` option."
        )
        return model

    def get_plan_edits_schema(
        self, plan: BaseModel | Any
    ) -> List[Type[BaseModel]]:
        schemas = []

        selections = self._extract_plan_selections(plan)
        if not selections:
            return schemas

        split_models = split_output_model_by_fields(self.builder.normalized)
        for field, model in split_models.items():
            if field in selections:
                schemas.append(model)

        return schemas

    def get_plan_edit_schema(self, plan: BaseModel) -> Type[BaseModel]:
        selections = self._extract_plan_selections(plan)
        schema = split_output_model(
            self.builder.normalized,  # type: ignore[arg-type]
            selections,
            partial=False,
        )

        doc = schema.__doc__
        if doc is None:
            doc = ""

        doc += (
            "Based on the edit plan you determined was appropriate, you have been provided a sparse representation"
            "of the fields representing the changes that should be made to the source data. Provide an edit for"
            "all fields provided."
        )
        schema.__doc__ = doc
        schema.__name__ = "Edits"

        return schema

    def _extract_plan_selections(self, plan: BaseModel) -> List[str]:
        selections = None
        if hasattr(plan, "selections"):
            selections = plan.selections
        elif hasattr(plan, "indices"):
            indices = plan.indices
            choices = getattr(plan.__class__, "_choices", None) or getattr(
                plan, "_choices", None
            )
            if choices is not None:
                selections = [choices[i] for i in indices]  # type: ignore[arg-type]
            else:
                selections = indices

        if selections in (None, "null"):
            return []

        if isinstance(selections, list):
            if "all" in selections:
                return self.builder.field_names
            if "null" in selections:
                return []
            return [s for s in selections if isinstance(s, str)]

        return []

    def apply_replacement(self, replacement: Any) -> Output:
        if isinstance(replacement, BaseModel):
            updates = replacement.model_dump(exclude_none=True)
        elif isinstance(replacement, dict):
            updates = {k: v for k, v in replacement.items() if v is not None}
        else:
            updates = replacement
        return self.builder.update(updates)

    def apply_selective(self, edits: Any) -> Output:
        if isinstance(edits, BaseModel) and hasattr(edits, "changes"):
            updates: dict[str, Any] = {}
            for change in edits.changes:
                field = getattr(change, "field", None)
                value = getattr(change, "value", None)
                if field is not None:
                    updates[field] = value
            return self.builder.update(updates)

        if isinstance(edits, dict):
            return self.builder.update(edits)
        return self.builder.update(edits)

    def apply_plan_edits(self, plan: BaseModel, edits: Any) -> Output:
        selections = self._extract_plan_selections(plan)
        if not selections:
            if self.builder.partial is None and self.builder.is_value:
                return self.builder.update(self.builder.target)
            return self.builder.partial  # type: ignore[return-value]

        if isinstance(edits, BaseModel):
            updates = edits.model_dump(exclude_none=True)
        elif isinstance(edits, dict):
            updates = {k: v for k, v in edits.items() if v is not None}
        else:
            updates = edits
        if selections:
            updates = {k: v for k, v in updates.items() if k in selections}
        return self.builder.update(updates)


@dataclass(init=False)
class EditStrategy(AbstractEditableStrategy[Output]):
    """Helper factory that auto-determines the appropriate edit strategy."""

    @property
    def kind(self) -> None:
        return None

    @classmethod
    def create(
        cls, builder: OutputBuilder[Output]
    ) -> AbstractEditableStrategy[Output]:
        base_text: str | None = None
        if isinstance(builder.partial, str):
            base_text = builder.partial
        elif builder.is_value and isinstance(builder.target, str):
            base_text = builder.target

        if builder.field_count > 0:
            return MappingEditStrategy(builder)

        if builder.normalized is str:
            if (
                base_text is not None
                and len(base_text) < SHORT_TEXT_REPLACEMENT_THRESHOLD
            ):
                return BasicEditStrategy(builder)
            return TextEditStrategy(builder)  # type: ignore[arg-type]

        return BasicEditStrategy(builder)



"""zyx._graph._nodes"""

from __future__ import annotations

from contextlib import AbstractAsyncContextManager
from dataclasses import replace
from typing import Any, List, TypeVar, Callable, Union, Generic, cast

from pydantic_graph.beta.step import StepContext
from pydantic_graph.nodes import End, GraphRunContext, BaseNode
from pydantic_ai.exceptions import UserError
from pydantic_ai.messages import DocumentUrl, ModelRequest, UserPromptPart

from .._aliases import (
    PydanticAIAgentResult,
    PydanticAIAgentStream,
)
from ._ctx import (
    SemanticGraphDeps,
    SemanticGraphState,
    StreamFieldMapping,
)
from ._requests import SemanticGraphRequestTemplate
from .._processing._multimodal import (
    MultimodalContentOrigin,
    MultimodalContentMediaType,
    render_multimodal_source_as_text,
)

__all__ = (
    "AbstractSemanticNode",
    "make_generate_step",
    "make_stream_step",
    "run_v1_node_chain",
)


# temporary workaround used if a model provider doesnt support native structured output
_NATIVE_OUTPUT_UNSUPPORTED_ERROR = "Native structured output is not supported"


Deps = TypeVar("Deps")
Output = TypeVar("Output")


class AbstractSemanticNode(
    Generic[Deps, Output],
    BaseNode[
        SemanticGraphState[Output], SemanticGraphDeps[Deps, Output], End | Any
    ],
):
    """Base class for v1-style semantic nodes used in edit workflows."""

    async def execute_run(
        self,
        ctx: GraphRunContext[
            SemanticGraphState[Output], SemanticGraphDeps[Deps, Output]
        ],
        *,
        request: SemanticGraphRequestTemplate[Output],
        update_output: bool = True,
        output_fields: str | List[str] | None = None,
    ) -> PydanticAIAgentResult[Output | Any]:
        result = await _execute_with_fallback(
            ctx,
            request=request,
            streaming=False,
        )

        if update_output:
            ctx.state.output.update_from_pydantic_ai_result(
                result=cast(PydanticAIAgentResult[Any], result),
                fields=output_fields,
            )

        ctx.state.agent_runs.append(result)  # type: ignore[arg-type]
        ctx.state.usage.incr(result.usage())  # type: ignore[arg-type]
        return result  # type: ignore[return-value]

    async def execute_stream(
        self,
        ctx: GraphRunContext[
            SemanticGraphState[Output], SemanticGraphDeps[Deps, Output]
        ],
        *,
        request: SemanticGraphRequestTemplate[Output],
    ) -> AbstractAsyncContextManager[PydanticAIAgentStream[Any, Output | Any]]:
        stream_ctx = await _execute_with_fallback(
            ctx,
            request=request,
            streaming=True,
        )
        assert isinstance(stream_ctx, AbstractAsyncContextManager)
        return stream_ctx  # type: ignore[return-value]


async def _execute_run(
    ctx: StepContext[
        SemanticGraphState[Output], SemanticGraphDeps[Deps, Output], Any
    ],
    *,
    request: SemanticGraphRequestTemplate[Output],
    update_output: bool = True,
    output_fields: str | List[str] | None = None,
) -> PydanticAIAgentResult[Output | Any]:
    result = await _execute_with_fallback(
        ctx,
        request=request,
        streaming=False,
    )

    if update_output:
        ctx.state.output.update_from_pydantic_ai_result(
            result=cast(PydanticAIAgentResult[Any], result),
            fields=output_fields,
        )

    ctx.state.agent_runs.append(result)  # type: ignore[arg-type]
    ctx.state.usage.incr(result.usage())  # type: ignore[arg-type]
    return result  # type: ignore[return-value]


async def _execute_with_fallback(
    ctx: Any,
    *,
    request: SemanticGraphRequestTemplate[Output],
    streaming: bool,
) -> Union[
    PydanticAIAgentResult[Output | Any],
    AbstractAsyncContextManager[PydanticAIAgentStream[Any, Output | Any]],
]:
    params = request.render(ctx)
    observe = getattr(ctx.deps, "observe", None)
    handler = (
        getattr(observe, "event_stream_handler", None) if observe else None
    )

    try:
        if observe and hasattr(observe, "on_model_request"):
            observe.on_model_request()
        return await _do_run(
            ctx,
            params=params,
            streaming=streaming,
            handler=handler,
            observe=observe,
        )
    except UserError as e:
        if _NATIVE_OUTPUT_UNSUPPORTED_ERROR not in str(e):
            raise
        fallback_request = replace(request, native_output=False)
        fallback_params = fallback_request.render(ctx)
        return await _do_run(
            ctx,
            params=fallback_params,
            streaming=streaming,
            handler=handler,
            observe=observe,
        )
    except RuntimeError as e:
        # TODO:
        # super janky fix, but it works for the time being.
        if "DocumentUrl is not supported" not in str(e):
            raise
        _inline_document_urls_in_params(params)
        return await _do_run(
            ctx,
            params=params,
            streaming=streaming,
            handler=handler,
            observe=observe,
        )


def _inline_document_urls_in_params(params: dict[str, Any]) -> None:
    message_history = params.get("message_history") or []
    for message in message_history:
        if not isinstance(message, ModelRequest):
            continue
        for part in message.parts:
            if not isinstance(part, UserPromptPart):
                continue
            content = part.content
            if not isinstance(content, list):
                continue
            new_content: list[Any] = []
            for item in content:
                if isinstance(item, DocumentUrl):
                    text = render_multimodal_source_as_text(
                        item.url,
                        MultimodalContentOrigin.URL,
                        MultimodalContentMediaType.DOCUMENT,
                    )
                    new_content.append(text)
                else:
                    new_content.append(item)
            part.content = new_content


async def _do_run(
    ctx: Any,
    *,
    params: dict[str, Any],
    streaming: bool,
    handler: Any,
    observe: Any,
) -> Union[
    PydanticAIAgentResult[Output | Any],
    AbstractAsyncContextManager[PydanticAIAgentStream[Any, Output | Any]],
]:
    if streaming:
        return ctx.deps.agent.run_stream(
            **params, event_stream_handler=handler
        )
    if observe:
        async with ctx.deps.agent.iter(**params) as agent_run:
            async for node in agent_run:
                observe.on_node(node, agent_run.ctx)
        result = agent_run.result
        if result is None:
            raise ValueError("Agent run finished without a result.")
        return result
    return await ctx.deps.agent.run(**params)


def make_generate_step(
    request: SemanticGraphRequestTemplate[Output],
    *,
    update_output: bool = True,
    output_fields: str | List[str] | None = None,
) -> Callable[
    [
        StepContext[
            SemanticGraphState[Output], SemanticGraphDeps[Deps, Output], Any
        ]
    ],
    Any,
]:
    async def _step(
        ctx: StepContext[
            SemanticGraphState[Output], SemanticGraphDeps[Deps, Output], Any
        ],
    ) -> Any:
        result = await _execute_run(
            ctx,
            request=request,
            update_output=update_output,
            output_fields=output_fields,
        )
        return result.output

    return _step


def make_stream_step(
    request: SemanticGraphRequestTemplate[Output],
    *,
    update_output: bool = True,
    output_fields: str | List[str] | None = None,
) -> Callable[
    [
        StepContext[
            SemanticGraphState[Output], SemanticGraphDeps[Deps, Output], Any
        ]
    ],
    Any,
]:
    async def _step(
        ctx: StepContext[
            SemanticGraphState[Output], SemanticGraphDeps[Deps, Output], Any
        ],
    ) -> Any:
        stream_ctx = await _execute_with_fallback(
            ctx,
            request=request,
            streaming=True,
        )
        assert isinstance(stream_ctx, AbstractAsyncContextManager)

        ctx.state.streams.append(None)
        ctx.state.stream_contexts.append(stream_ctx)

        if output_fields:
            fields = (
                output_fields
                if isinstance(output_fields, list)
                else [output_fields]
            )
            ctx.state.stream_field_mappings.append(
                StreamFieldMapping(
                    stream_index=len(ctx.state.streams) - 1,
                    fields=fields,
                    update_output=update_output,
                )
            )
        else:
            ctx.state.stream_field_mappings.append(
                StreamFieldMapping(
                    stream_index=len(ctx.state.streams) - 1,
                    fields=None,
                    update_output=update_output,
                )
            )

        return None

    return _step


async def run_v1_node_chain(
    start_node: BaseNode[Any, Any, Any],
    ctx: StepContext[
        SemanticGraphState[Output], SemanticGraphDeps[Deps, Output], Any
    ],
) -> Any:
    node: BaseNode[Any, Any, Any] = start_node
    while True:
        result = await node.run(
            GraphRunContext(state=ctx.state, deps=ctx.deps)
        )
        if isinstance(result, End):
            return result.data
        if not isinstance(result, BaseNode):
            raise ValueError(f"Invalid node transition: {type(result)}")
        node = result



"""zyx.processing.multimodal"""

from __future__ import annotations

from functools import lru_cache
from enum import Enum
from io import BytesIO
import mimetypes
import re
from urllib.parse import urlparse

import httpx
import urllib.request
import urllib.error
from pathlib import Path
from typing import Any, TypeAlias, Self, Tuple, TYPE_CHECKING

from pydantic_ai.messages import (
    AudioUrl,
    BinaryContent,
    DocumentUrl,
    ImageUrl,
    VideoUrl,
    UserContent,
)

from ._toon import object_as_text

if TYPE_CHECKING:
    from markitdown import MarkItDown

__all__ = (
    "MultimodalContent",
    "MultimodalContentOrigin",
    "MultimodalContentMediaType",
    "classify_multimodal_source",
    "render_multimodal_source_as_user_content",
    "render_multimodal_source_as_text",
    "render_multimodal_source_as_description",
)


MultimodalContent: TypeAlias = Path | str | bytes | Any
"""
The source location from where a piece of external content can originate from.

This can be one of the following types:
- A `Path` object representing a local file
- A string representing a URL to a text or multimodal file/html page
- A bytes object representing raw binary data
- Any non-string python object that can be encoded to text using the TOON format
"""


_URIS = ("http://", "https://", "ftp://")
_TEXT_EXTENSIONS: frozenset[str] = frozenset(
    (".txt", ".md", ".rst", ".json", ".yaml", ".yml", ".ini", ".toml", ".log")
)
_DOCUMENT_MIMETYPES: frozenset[str] = frozenset(
    (
        "application/pdf",
        "application/msword",
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
        "application/vnd.ms-powerpoint",
        "application/vnd.openxmlformats-officedocument.presentationml.presentation",
        "application/vnd.ms-excel",
        "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    )
)
_URL_HEADERS = {
    "User-Agent": "zyx/1.0",
    "Accept": "*/*",
}


def _looks_like_url(value: str) -> bool:
    if not value or " " in value:
        return False
    if value.startswith(_URIS):
        return True
    # Avoid misclassifying local filesystem paths like "foo/bar.txt"
    if Path(value).exists():
        return False
    candidate = _normalize_url(value)
    try:
        parsed = httpx.URL(candidate)
    except Exception:
        return False
    host = parsed.host
    if not host:
        return False
    if host == "localhost":
        return True
    return "." in host
    return True


def _normalize_url(value: str) -> str:
    if value.startswith(_URIS):
        return value
    parsed = urlparse(value)
    if parsed.scheme:
        return value
    return f"https://{value}"


class MultimodalContentOrigin(Enum):
    """
    The computed origin of a snippet.
    """

    FILE = "file"
    URL = "url"
    STRING = "string"
    BYTES = "bytes"
    OBJECT = "object"

    @classmethod
    def classify(cls, source: MultimodalContent) -> Self:
        """
        Classifies the the origin of a snippet's source.
        """
        if isinstance(source, (bytes, bytearray, memoryview)):
            return cls.BYTES  # type: ignore

        elif isinstance(source, Path):
            if source.is_file():
                return cls.FILE  # type: ignore

        else:
            if isinstance(source, str):
                if source.startswith(_URIS):
                    return cls.URL  # type: ignore
                if _looks_like_url(source):
                    return cls.URL  # type: ignore
                if Path(source).exists() and Path(source).is_file():
                    return cls.FILE  # type: ignore

            else:
                return cls.OBJECT  # type: ignore

        return cls.STRING  # type: ignore


class MultimodalContentMediaType(Enum):
    """
    The media/content type of a source.
    """

    TEXT = "text/plain"
    HTML = "text/html"
    DOCUMENT = "document"
    IMAGE = "image"
    AUDIO = "audio"
    VIDEO = "video"
    UNKNOWN = "unknown"

    @classmethod
    def classify(
        cls, source: MultimodalContent, origin: MultimodalContentOrigin
    ) -> Self:
        """
        Classifies the the type of a snippet's content.
        """
        if (
            origin == MultimodalContentOrigin.OBJECT
            or origin == MultimodalContentOrigin.STRING
        ):
            # objects are encoded to text using the TOON format
            return cls.TEXT  # type: ignore

        if origin == MultimodalContentOrigin.BYTES:
            data = (
                source.tobytes()
                if isinstance(source, memoryview)
                else bytes(source)  # type: ignore
            )
            return _classify_magic_bytes(data)  # type: ignore

        path_text = ""
        suffix = ""
        if origin == MultimodalContentOrigin.FILE:
            path = source if isinstance(source, Path) else Path(source)  # type: ignore
            path_text = str(path)
            suffix = path.suffix.lower()
        elif origin == MultimodalContentOrigin.URL and isinstance(source, str):
            path_text = re.split(r"[?#]", source, 1)[0]
            suffix = Path(path_text).suffix.lower()

        if suffix in _TEXT_EXTENSIONS:
            return cls.TEXT  # type: ignore
        if suffix in (".html", ".htm", ".xhtml"):
            return cls.HTML  # type: ignore
        if suffix in (
            ".pdf",
            ".doc",
            ".docx",
            ".ppt",
            ".pptx",
            ".xls",
            ".xlsx",
        ):
            return cls.DOCUMENT  # type: ignore

        mime_type, _ = mimetypes.guess_type(path_text)
        if mime_type == "text/html":
            return cls.HTML  # type: ignore
        if mime_type in _DOCUMENT_MIMETYPES:
            return cls.DOCUMENT  # type: ignore
        if mime_type:
            if mime_type.startswith("image/"):
                return cls.IMAGE  # type: ignore
            if mime_type.startswith("audio/"):
                return cls.AUDIO  # type: ignore
            if mime_type.startswith("video/"):
                return cls.VIDEO  # type: ignore
            if mime_type.startswith("text/"):
                return cls.TEXT  # type: ignore

        if (
            origin == MultimodalContentOrigin.URL
            and not suffix
            and not mime_type
        ):
            detected, _ = _detect_url_media_type(source)
            if detected:
                return detected  # type: ignore
            # Default unknown URLs to text to avoid treating common web pages as UNKNOWN.
            return cls.TEXT  # type: ignore

        return cls.UNKNOWN  # type: ignore


def _classify_content_type(
    content_type: str | None,
) -> MultimodalContentMediaType | None:
    if not content_type:
        return None
    ct = content_type.split(";", 1)[0].strip().lower()
    if not ct:
        return None
    if ct == "text/html":
        return MultimodalContentMediaType.HTML
    if ct in _DOCUMENT_MIMETYPES or ct == "application/pdf":
        return MultimodalContentMediaType.DOCUMENT
    if ct.startswith("image/"):
        return MultimodalContentMediaType.IMAGE
    if ct.startswith("audio/"):
        return MultimodalContentMediaType.AUDIO
    if ct.startswith("video/"):
        return MultimodalContentMediaType.VIDEO
    if ct.startswith("text/"):
        return MultimodalContentMediaType.TEXT
    return None


def _classify_magic_bytes(
    data: bytes,
) -> MultimodalContentMediaType | None:
    head = data[:512]
    if head.startswith(b"%PDF"):
        return MultimodalContentMediaType.DOCUMENT
    if head.startswith(b"\x89PNG\r\n\x1a\n") or head[:3] == b"\xff\xd8\xff":
        return MultimodalContentMediaType.IMAGE
    if head.startswith(b"GIF87a") or head.startswith(b"GIF89a"):
        return MultimodalContentMediaType.IMAGE
    if head[:4] == b"RIFF" and head[8:12] == b"WEBP":
        return MultimodalContentMediaType.IMAGE
    if head[:4] == b"RIFF" and head[8:12] == b"WAVE":
        return MultimodalContentMediaType.AUDIO
    if (
        head.startswith(b"OggS")
        or head.startswith(b"fLaC")
        or head.startswith(b"ID3")
    ):
        return MultimodalContentMediaType.AUDIO
    if len(head) >= 8 and head[4:8] == b"ftyp":
        return MultimodalContentMediaType.VIDEO
    if head.startswith(b"\x1a\x45\xdf\xa3"):
        return MultimodalContentMediaType.VIDEO
    if b"\x00" in head:
        return MultimodalContentMediaType.UNKNOWN
    try:
        text = head.decode("utf-8")
    except UnicodeDecodeError:
        return MultimodalContentMediaType.UNKNOWN
    if re.search(r"(?i)<!doctype html|<html\\b|<head\\b|<body\\b", text):
        return MultimodalContentMediaType.HTML
    return MultimodalContentMediaType.TEXT


@lru_cache(maxsize=128)
def _detect_url_media_type(
    url: str,
) -> Tuple[MultimodalContentMediaType | None, str | None]:
    url = _normalize_url(url)
    try:
        req = urllib.request.Request(url, method="HEAD", headers=_URL_HEADERS)
        with urllib.request.urlopen(req, timeout=5) as resp:
            content_type = resp.headers.get("Content-Type")
            classified = _classify_content_type(content_type)
            return classified, (
                content_type.split(";", 1)[0].strip().lower()
                if content_type
                else None
            )
    except (urllib.error.URLError, urllib.error.HTTPError):
        pass

    try:
        req = urllib.request.Request(
            url,
            method="GET",
            headers={**_URL_HEADERS, "Range": "bytes=0-511"},
        )
        with urllib.request.urlopen(req, timeout=5) as resp:
            content_type = resp.headers.get("Content-Type")
            ct_class = _classify_content_type(content_type)
            if ct_class:
                return ct_class, (
                    content_type.split(";", 1)[0].strip().lower()
                    if content_type
                    else None
                )
            data = resp.read(512)
            return _classify_magic_bytes(data), None
    except (urllib.error.URLError, urllib.error.HTTPError):
        return None, None


def classify_multimodal_source(
    source: MultimodalContent,
) -> Tuple[MultimodalContentOrigin, MultimodalContentMediaType]:
    """
    Classifies a multimodal source into its origin and type.
    """
    origin = MultimodalContentOrigin.classify(source)
    return origin, MultimodalContentMediaType.classify(source, origin)


@lru_cache(maxsize=1)
def _get_markdown_converter() -> MarkItDown:
    from markitdown import MarkItDown

    return MarkItDown()


def _render_multimodal_file_or_url_as_text(source: str) -> str:
    try:
        return _get_markdown_converter().convert(source).markdown
    except Exception as e:
        raise ValueError(f"Failed to render file or URL as text: {e}") from e


def render_multimodal_source_as_user_content(
    source: MultimodalContent,
    origin: MultimodalContentOrigin,
    media_type: MultimodalContentMediaType,
) -> UserContent:
    """
    Renders a multimodal source as a compatible `pydantic-ai` UserContent object.
    """
    if origin == MultimodalContentOrigin.OBJECT:
        return object_as_text(source)

    if origin == MultimodalContentOrigin.STRING:
        if media_type == MultimodalContentMediaType.HTML:
            from markitdown import StreamInfo

            buffer = BytesIO(str(source).encode("utf-8"))
            info = StreamInfo(mimetype="text/html", extension=".html")
            return (
                _get_markdown_converter()
                .convert(buffer, stream_info=info)
                .markdown
            )
        return str(source)

    if origin == MultimodalContentOrigin.URL and isinstance(source, str):
        normalized = _normalize_url(source)
        guessed_media_type, _ = mimetypes.guess_type(normalized)
        if not guessed_media_type:
            _, detected = _detect_url_media_type(normalized)
            guessed_media_type = detected
        if media_type == MultimodalContentMediaType.IMAGE:
            return ImageUrl(url=normalized, media_type=guessed_media_type)
        if media_type == MultimodalContentMediaType.AUDIO:
            return AudioUrl(url=normalized, media_type=guessed_media_type)
        if media_type == MultimodalContentMediaType.VIDEO:
            return VideoUrl(url=normalized, media_type=guessed_media_type)
        if media_type in (
            MultimodalContentMediaType.TEXT,
            MultimodalContentMediaType.HTML,
            MultimodalContentMediaType.DOCUMENT,
        ):
            return _render_multimodal_file_or_url_as_text(normalized)
        return (
            DocumentUrl(url=normalized, media_type=guessed_media_type)
            if guessed_media_type
            else normalized
        )

    if origin == MultimodalContentOrigin.FILE:
        path = source if isinstance(source, Path) else Path(source)  # type: ignore
        if media_type in (
            MultimodalContentMediaType.TEXT,
            MultimodalContentMediaType.HTML,
            MultimodalContentMediaType.DOCUMENT,
        ):
            return _render_multimodal_file_or_url_as_text(str(path))
        guessed_media_type, _ = mimetypes.guess_type(str(path))
        data = path.read_bytes()
        return BinaryContent.narrow_type(
            BinaryContent(
                data=data,
                media_type=guessed_media_type or "application/octet-stream",
            )
        )

    if origin == MultimodalContentOrigin.BYTES:
        data = (
            source.tobytes()
            if isinstance(source, memoryview)
            else bytes(source)  # type: ignore
        )
        if media_type == MultimodalContentMediaType.HTML:
            from markitdown import StreamInfo

            buffer = BytesIO(data)
            info = StreamInfo(mimetype="text/html", extension=".html")
            return (
                _get_markdown_converter()
                .convert(buffer, stream_info=info)
                .markdown
            )
        if media_type == MultimodalContentMediaType.TEXT:
            return data.decode("utf-8")
        return BinaryContent.narrow_type(
            BinaryContent(data=data, media_type="application/octet-stream")
        )

    return str(source)


def render_multimodal_source_as_text(
    source: MultimodalContent,
    origin: MultimodalContentOrigin,
    media_type: MultimodalContentMediaType,
) -> str:
    """
    Renders a multimodal source as text, or a textual description for
    multimodal content.
    """
    if origin == MultimodalContentOrigin.OBJECT:
        return object_as_text(source)

    if media_type in (
        MultimodalContentMediaType.TEXT,
        MultimodalContentMediaType.HTML,
        MultimodalContentMediaType.DOCUMENT,
        MultimodalContentMediaType.UNKNOWN,
    ):
        if origin == MultimodalContentOrigin.STRING and isinstance(
            source, str
        ):
            return source

        if origin == MultimodalContentOrigin.FILE:
            path = source if isinstance(source, Path) else Path(source)  # type: ignore
            return _render_multimodal_file_or_url_as_text(str(path))

        if origin == MultimodalContentOrigin.URL and isinstance(source, str):
            return _render_multimodal_file_or_url_as_text(
                _normalize_url(source)
            )

        if origin == MultimodalContentOrigin.BYTES:
            data = (
                source.tobytes()
                if isinstance(source, memoryview)
                else bytes(source)  # type: ignore
            )
            if media_type == MultimodalContentMediaType.TEXT:
                return data.decode("utf-8")
            if media_type == MultimodalContentMediaType.HTML:
                from markitdown import StreamInfo

                buffer = BytesIO(data)
                info = StreamInfo(mimetype="text/html", extension=".html")
                return (
                    _get_markdown_converter()
                    .convert(buffer, stream_info=info)
                    .markdown
                )
            return "[document bytes]"

    if origin == MultimodalContentOrigin.URL and isinstance(source, str):
        return f"[{media_type.value} content at {_normalize_url(source)}]"
    if origin == MultimodalContentOrigin.FILE:
        path = source if isinstance(source, Path) else Path(source)  # type: ignore
        return f"[{media_type.value} content from {path}]"
    if origin == MultimodalContentOrigin.BYTES:
        return f"[{media_type.value} content bytes]"

    return str(source)


def render_multimodal_source_as_description(
    source: MultimodalContent,
    origin: MultimodalContentOrigin,
    media_type: MultimodalContentMediaType,
) -> str:
    """
    Renders a multimodal source as a description, or a textual description for
    multimodal content.
    """
    if origin == MultimodalContentOrigin.OBJECT:
        return f"[object of type {type(source).__name__}]"

    if origin == MultimodalContentOrigin.STRING:
        length = len(source) if isinstance(source, str) else 0
        return f"[string value, length={length}]"

    if origin == MultimodalContentOrigin.URL:
        return f"[URL: {source} | type={media_type.value}]"

    if origin == MultimodalContentOrigin.FILE:
        path = source if isinstance(source, Path) else Path(source)  # type: ignore[arg-type]
        name = path.name or "file"
        return f"[{media_type.value} file: {name}]"

    if origin == MultimodalContentOrigin.BYTES:
        size = (
            len(source)
            if isinstance(source, (bytes, bytearray))
            else len(bytes(source))  # type: ignore
        )
        return f"[{media_type.value} bytes, size={size}]"

    return f"[{media_type.value} content]"
